from __future__ import annotations

"""
Niquests API Client
"""


import atexit
import typing as t
from lzl import load
from urllib.parse import urljoin
from .utils import logger

if load.TYPE_CHECKING:
    import niquests
    from niquests import Session, AsyncSession, Response, AsyncResponse
    from niquests.sessions import ResolverType, HttpAuthenticationType, HttpMethodType, QueryParameterType, BodyType, HeadersType, CookiesType, MultiPartFilesType, MultiPartFilesAltType, TimeoutType, ProxyType, HookType, TLSVerifyType, TLSClientCertType
    from niquests.sessions import PreparedRequest, BaseAdapter
    from niquests._async import AsyncResolverType, CacheLayerAltSvcType, RetryType
    from niquests._typing import AsyncHttpAuthenticationType, AsyncBodyType
    from niquests._async import AsyncBaseAdapter
else:
    niquests = load.LazyLoad("niquests", install_missing = True)


DEFAULT_POOLBLOCK: bool = False
DEFAULT_POOLSIZE: int = 10
DEFAULT_RETRIES: 'RetryType' = 0
DEFAULT_TIMEOUT: 'TimeoutType' = 120


class Client:
    """A Unified Requests Client.

    Provides cookie persistence, connection-pooling, and configuration.

    Basic Usage::

      >>> import niquests
      >>> s = niquests.AsyncSession()
      >>> await s.get('https://httpbin.org/get')
      <Response HTTP/2 [200]>

    Or as a context manager::

      >>> async with niquests.AsyncSession() as s:
      ...     await s.get('https://httpbin.org/get')
      <Response HTTP/2 [200]>
    """


    def __init__(
        self,
        *,
        base_url: t.Optional[str] = None,
        headers: t.Optional[t.Dict[str, str]] = None,
        auth: t.Optional['HttpAuthenticationType'] = None,
        proxies: 'ProxyType' | None = None,
        cookies: 'CookiesType' | None = None,
        verify: 'TLSVerifyType' = None,
        allow_redirects: bool = None,
        timeout: 'TimeoutType' = None,
        cert: 'TLSClientCertType' | None = None,
        hooks: 'HookType[PreparedRequest | Response]' | None = None,
        ahooks: 'HookType[PreparedRequest | Response]' | None = None,

        resolver: 'AsyncResolverType' | 'ResolverType' | None = None, 
        source_address: tuple[str, int] | None = None,
        quic_cache_layer: 'CacheLayerAltSvcType' | None = None,
        retries: 'RetryType' = DEFAULT_RETRIES,
        multiplexed: bool = False,
        disable_http1: bool = False,
        disable_http2: bool = False,
        disable_http3: bool = False,
        disable_ipv6: bool = False,
        disable_ipv4: bool = False,
        pool_connections: int = DEFAULT_POOLSIZE,
        pool_maxsize: int = DEFAULT_POOLSIZE,
        happy_eyeballs: bool | int = False,
        auto_close_on_exit: t.Optional[bool] = True,
        **kwargs,
    ):
        """
        Initialize a new Client. This is a wrapper around `niquests.Session` and `niquests.AsyncSession`.

        It allows setting intial headers, auth, proxies, cookies, verify, allow_redirects, timeout, and cert.
        that are used for all requests.

        :param base_url: The base url to append to each url in the request.
        :param headers: The headers to use for all requests.
        :param auth: The auth to use for all requests.
        :param proxies: The proxies to use for all requests.
        :param cookies: The cookies to use for all requests.
        :param verify: The verify to use for all requests.
        :param allow_redirects: The allow_redirects to use for all requests.
        :param timeout: The timeout to use for all requests.
        :param cert: The cert to use for all requests.
        :param hooks: The hooks to use for all sync requests.
        :param ahooks: The hooks to use for all async requests.

        :param resolver: Specify a DNS resolver that should be used within this Session.
        :param source_address: Bind Session to a specific network adapter and/or port so that all outgoing requests.
        :param quic_cache_layer: Provide an external cache mechanism to store HTTP/3 host capabilities.
        :param retries: Configure a number of times a request must be automatically retried before giving up.
        :param multiplexed: Enable or disable concurrent request when the remote host support HTTP/2 onward.
        :param disable_http2: Toggle to disable negotiating HTTP/2 with remote peers.
        :param disable_http3: Toggle to disable negotiating HTTP/3 with remote peers.
        :param disable_ipv6: Toggle to disable using IPv6 even if the remote host supports IPv6.
        :param disable_ipv4: Toggle to disable using IPv4 even if the remote host supports IPv4.
        :param pool_connections: Number of concurrent hosts to be handled by this Session at a maximum.
        :param pool_maxsize: Maximum number of concurrent connections per (single) host at a time.

        """
        self.base_url = base_url
        self._base_headers = headers or {}
        self._base_auth = auth
        self._base_proxies = proxies
        self._base_cookies = cookies
        self._base_verify = verify
        if 'follow_redirects' in kwargs: allow_redirects = kwargs.pop('follow_redirects')
        
        self._base_allow_redirects = allow_redirects
        self._base_timeout = timeout
        self._base_cert = cert
        self._base_hooks = hooks
        self._base_ahooks = ahooks

        self._auto_close_on_exit = auto_close_on_exit

        self._io: t.Optional['Session'] = None
        self._aio: t.Optional['AsyncSession'] = None
        self._extra: t.Dict[str, t.Any] = {}
        self._client_kwargs: t.Dict[str, t.Any] = {
            'resolver': resolver,
            'source_address': source_address,
            'quic_cache_layer': quic_cache_layer,
            'retries': retries,
            'multiplexed': multiplexed,
            'disable_http1': disable_http1,
            'disable_http2': disable_http2,
            'disable_http3': disable_http3,
            'disable_ipv6': disable_ipv6,
            'disable_ipv4': disable_ipv4,
            'pool_connections': pool_connections,
            'pool_maxsize': pool_maxsize,
            'happy_eyeballs': happy_eyeballs,
        }

    def _get_url_(self, url: str) -> str:
        """
        Returns the URL
        """
        return urljoin(self.base_url, url) if self.base_url else url
    
    def _get_request_kwargs_(self, method: str, **kwargs) -> t.Dict[str, t.Any]:
        """
        Returns the request kwargs
        """
        method = method.upper()
        if 'cert' not in kwargs and self._base_cert:
            kwargs['cert'] = self._base_cert
        if 'timeout' not in kwargs and self._base_timeout is not None:
            kwargs['timeout'] = self._base_timeout
        if 'allow_redirects' not in kwargs and self._base_allow_redirects is not None:
            kwargs['allow_redirects'] = self._base_allow_redirects
        return kwargs


    def _update_session_(self, session: t.Union['Session', 'AsyncSession']) -> t.Union['Session', 'AsyncSession']:
        # sourcery skip: merge-else-if-into-elif
        """
        Updates the session
        """
        if self._base_headers: session.headers.update(self._base_headers)
        if self._base_auth: session.auth = self._base_auth
        if self._base_proxies: session.proxies.update(self._base_proxies)
        if self._base_cookies: session.cookies.update(self._base_cookies)
        if self._base_verify is not None: session.verify = self._base_verify
        # This is the async session
        if hasattr(session, '__aenter__'):
            if self._base_ahooks: session.hooks.update(self._base_ahooks)
        else:
            if self._base_hooks: session.hooks.update(self._base_hooks)
        return session
        
    
    @property
    def io(self) -> 'Session':
        """
        Returns the io session
        """
        if self._io is None:
            self._io = niquests.Session(**self._client_kwargs)
            self._io = self._update_session_(self._io)
            if self._auto_close_on_exit:
                atexit.register(self.close)
        return self._io
    
    @property
    def aio(self) -> 'AsyncSession':
        """
        Returns the async session
        """
        if self._aio is None:
            self._aio = niquests.AsyncSession(**self._client_kwargs)
            self._aio = self._update_session_(self._aio)
            if self._auto_close_on_exit:
                from lzo.utils import aioexit
                aioexit.register(self.aclose)
        return self._aio
    
    def close(self):
        """
        Closes the client
        """
        if self._io is not None: 
            self._io.close()
            self._io = None

    async def aclose(self):
        """
        Closes the client
        """
        self.close()
        if self._aio is not None: 
            await self._aio.close()
            self._aio = None
    
    """
    Sync Client Methods
    """

    @t.overload
    def request(
        self,
        method: 'HttpMethodType',
        url: str,
        params: 'QueryParameterType' | None = ...,
        data: 'BodyType' | None = ...,
        headers: 'HeadersType' | None = ...,
        cookies: 'CookiesType' | None = ...,
        files: 'MultiPartFilesType' | 'MultiPartFilesAltType' | None = ...,
        auth: 'HttpAuthenticationType' | None = ...,
        timeout: 'TimeoutType' | None = ...,
        allow_redirects: bool = True,
        proxies: 'ProxyType' | None = ...,
        hooks: 'HookType[PreparedRequest | Response]' | None = ...,
        stream: bool | None = ...,
        verify: 'TLSVerifyType' | None = ...,
        cert: 'TLSClientCertType' | None = ...,
        json: t.Any | None = ...,
    ) -> 'Response':
        """Constructs a :class:`Request <Request>`, prepares it and sends it.
        Returns :class:`Response <Response>` object.

        :param method: method for the new :class:`Request` object.
        :param url: URL for the new :class:`Request` object.
        :param params: (optional) Dictionary or bytes to be sent in the query
            string for the :class:`Request`.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param json: (optional) json to send in the body of the
            :class:`Request`.
        :param headers: (optional) Dictionary of HTTP Headers to send with the
            :class:`Request`.
        :param cookies: (optional) Dict or CookieJar object to send with the
            :class:`Request`.
        :param files: (optional) Dictionary of ``'filename': file-like-objects``
            for multipart encoding upload.
        :param auth: (optional) Auth tuple or callable to enable
            Basic/Digest/Custom HTTP Auth.
        :param timeout: (optional) How long to wait for the server to send
            data before giving up, as a float, or a :ref:`(connect timeout,
            read timeout) <timeouts>` tuple.
        :param allow_redirects: (optional) Set to True by default.
        :param proxies: (optional) Dictionary mapping protocol or protocol and
            hostname to the URL of the proxy.
        :param hooks: (optional) Dictionary mapping hook name to one event or
            list of events, event must be callable.
        :param stream: (optional) whether to immediately download the response
            content. Defaults to ``False``.
        :param verify: (optional) Either a boolean, in which case it controls whether we verify
            the server's TLS certificate, or a string, in which case it must be a path
            to a CA bundle to use. Defaults to ``True``. When set to
            ``False``, requests will accept any TLS certificate presented by
            the server, and will ignore hostname mismatches and/or expired
            certificates, which will make your application vulnerable to
            man-in-the-middle (MitM) attacks. Setting verify to ``False``
            may be useful during local development or testing.
            It is also possible to put the certificates (directly) in a string or bytes.
        :param cert: (optional) if String, path to ssl client cert file (.pem).
            If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
        """
        ...
    
    def request(
        self,
        method: 'HttpMethodType',
        url: str,
        **kwargs,
    ) -> 'Response':
        url = self._get_url_(url)
        kwargs = self._get_request_kwargs_(method, **kwargs)
        return self.io.request(method, url, **kwargs)

    @t.overload
    def get(
        self,
        url: str,
        *,
        params: 'QueryParameterType' | None = ...,
        headers: 'HeadersType' | None = ...,
        cookies: 'CookiesType' | None = ...,
        auth: 'HttpAuthenticationType' | None = ...,
        timeout: 'TimeoutType' | None = ...,
        allow_redirects: bool = ...,
        proxies: 'ProxyType' | None = ...,
        hooks: 'HookType[PreparedRequest | Response]' | None = ...,
        verify: 'TLSVerifyType' = ...,
        stream: bool = ...,
        cert: 'TLSClientCertType' | None = ...,
        **kwargs: t.Any,
    ) -> 'Response':
        r"""Sends a GET request. Returns :class:`Response` object.

        :param url: URL for the new :class:`Request` object.
        :param params: (optional) Dictionary or bytes to be sent in the query
            string for the :class:`Request`.
        :param headers: (optional) Dictionary of HTTP Headers to send with the
            :class:`Request`.
        :param cookies: (optional) Dict or CookieJar object to send with the
            :class:`Request`.
        :param auth: (optional) Auth tuple or callable to enable
            Basic/Digest/Custom HTTP Auth.
        :param timeout: (optional) How long to wait for the server to send
            data before giving up, as a float, or a :ref:`(connect timeout,
            read timeout) <timeouts>` tuple.
        :param allow_redirects: (optional) Set to True by default.
        :param proxies: (optional) Dictionary mapping protocol or protocol and
            hostname to the URL of the proxy.
        :param hooks: (optional) Dictionary mapping hook name to one event or
            list of events, event must be callable.
        :param stream: (optional) whether to immediately download the response
            content. Defaults to ``False``.
        :param verify: (optional) Either a boolean, in which case it controls whether we verify
            the server's TLS certificate, or a string, in which case it must be a path
            to a CA bundle to use. Defaults to ``True``. When set to
            ``False``, requests will accept any TLS certificate presented by
            the server, and will ignore hostname mismatches and/or expired
            certificates, which will make your application vulnerable to
            man-in-the-middle (MitM) attacks. Setting verify to ``False``
            may be useful during local development or testing.
            It is also possible to put the certificates (directly) in a string or bytes.
        :param cert: (optional) if String, path to ssl client cert file (.pem).
            If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
        """
        ...
    
    def get(
        self,
        url: str,
        *args,
        **kwargs,
    ) -> 'Response':
        return self.request(
            "GET",
            url,
            **kwargs,
        )

    @t.overload
    def options(
        self,
        url: str,
        *,
        params: 'QueryParameterType' | None = ...,
        headers: 'HeadersType' | None = ...,
        cookies: 'CookiesType' | None = ...,
        auth: 'HttpAuthenticationType' | None = ...,
        timeout: 'TimeoutType' | None = ...,
        allow_redirects: bool = True,
        proxies: 'ProxyType' | None = ...,
        hooks: 'HookType[PreparedRequest | Response]' | None = ...,
        verify: 'TLSVerifyType' = ...,
        stream: bool = ...,
        cert: 'TLSClientCertType' | None = ...,
        **kwargs: t.Any,
    ) -> 'Response':
        r"""Sends a OPTIONS request. Returns :class:`Response` object.

        :param url: URL for the new :class:`Request` object.
        :param params: (optional) Dictionary or bytes to be sent in the query
            string for the :class:`Request`.
        :param headers: (optional) Dictionary of HTTP Headers to send with the
            :class:`Request`.
        :param cookies: (optional) Dict or CookieJar object to send with the
            :class:`Request`.
        :param auth: (optional) Auth tuple or callable to enable
            Basic/Digest/Custom HTTP Auth.
        :param timeout: (optional) How long to wait for the server to send
            data before giving up, as a float, or a :ref:`(connect timeout,
            read timeout) <timeouts>` tuple.
        :param allow_redirects: (optional) Set to True by default.
        :param proxies: (optional) Dictionary mapping protocol or protocol and
            hostname to the URL of the proxy.
        :param hooks: (optional) Dictionary mapping hook name to one event or
            list of events, event must be callable.
        :param stream: (optional) whether to immediately download the response
            content. Defaults to ``False``.
        :param verify: (optional) Either a boolean, in which case it controls whether we verify
            the server's TLS certificate, or a string, in which case it must be a path
            to a CA bundle to use. Defaults to ``True``. When set to
            ``False``, requests will accept any TLS certificate presented by
            the server, and will ignore hostname mismatches and/or expired
            certificates, which will make your application vulnerable to
            man-in-the-middle (MitM) attacks. Setting verify to ``False``
            may be useful during local development or testing.
            It is also possible to put the certificates (directly) in a string or bytes.
        :param cert: (optional) if String, path to ssl client cert file (.pem).
            If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
        """
        ...
    
    def options(
        self,
        url: str,
        *args,
        **kwargs,
    ) -> 'Response':
        return self.request(
            "OPTIONS",
            url,
            **kwargs,
        )

    @t.overload
    def head(
        self,
        url: str,
        *,
        params: 'QueryParameterType' | None = ...,
        headers: 'HeadersType' | None = ...,
        cookies: 'CookiesType' | None = ...,
        auth: 'HttpAuthenticationType' | None = ...,
        timeout: 'TimeoutType' | None = ...,
        allow_redirects: bool = True,
        proxies: 'ProxyType' | None = ...,
        hooks: 'HookType[PreparedRequest | Response]' | None = ...,
        verify: 'TLSVerifyType' = ...,
        stream: bool = ...,
        cert: 'TLSClientCertType' | None = ...,
        **kwargs: t.Any,
    ) -> 'Response':
        r"""Sends a HEAD request. Returns :class:`Response` object.

        :param url: URL for the new :class:`Request` object.
        :param params: (optional) Dictionary or bytes to be sent in the query
            string for the :class:`Request`.
        :param headers: (optional) Dictionary of HTTP Headers to send with the
            :class:`Request`.
        :param cookies: (optional) Dict or CookieJar object to send with the
            :class:`Request`.
        :param auth: (optional) Auth tuple or callable to enable
            Basic/Digest/Custom HTTP Auth.
        :param timeout: (optional) How long to wait for the server to send
            data before giving up, as a float, or a :ref:`(connect timeout,
            read timeout) <timeouts>` tuple.
        :param allow_redirects: (optional) Set to False by default.
        :param proxies: (optional) Dictionary mapping protocol or protocol and
            hostname to the URL of the proxy.
        :param hooks: (optional) Dictionary mapping hook name to one event or
            list of events, event must be callable.
        :param stream: (optional) whether to immediately download the response
            content. Defaults to ``False``.
        :param verify: (optional) Either a boolean, in which case it controls whether we verify
            the server's TLS certificate, or a string, in which case it must be a path
            to a CA bundle to use. Defaults to ``True``. When set to
            ``False``, requests will accept any TLS certificate presented by
            the server, and will ignore hostname mismatches and/or expired
            certificates, which will make your application vulnerable to
            man-in-the-middle (MitM) attacks. Setting verify to ``False``
            may be useful during local development or testing.
            It is also possible to put the certificates (directly) in a string or bytes.
        :param cert: (optional) if String, path to ssl client cert file (.pem).
            If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
        """
        ...
    
    def head(
        self,
        url: str,
        *args,
        **kwargs,
    ) -> 'Response':
        return self.request(
            "HEAD",
            url,
            **kwargs,
        )

    @t.overload
    def post(
        self,
        url: str,
        data: 'BodyType' | None = ...,
        json: t.Any | None = ...,
        *,
        params: 'QueryParameterType' | None = ...,
        headers: 'HeadersType' | None = ...,
        cookies: 'CookiesType' | None = ...,
        files: 'MultiPartFilesType' | 'MultiPartFilesAltType' | None = ...,
        auth: 'HttpAuthenticationType' | None = ...,
        timeout: 'TimeoutType' | None = ...,
        allow_redirects: bool = ...,
        proxies: 'ProxyType' | None = ...,
        hooks: 'HookType[PreparedRequest | Response]' | None = ...,
        verify: 'TLSVerifyType' = ...,
        stream: bool = ...,
        cert: 'TLSClientCertType' | None = ...,
    ) -> 'Response':
        r"""Sends a POST request. Returns :class:`Response` object.

        :param url: URL for the new :class:`Request` object.
        :param params: (optional) Dictionary or bytes to be sent in the query
            string for the :class:`Request`.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param json: (optional) json to send in the body of the
            :class:`Request`.
        :param headers: (optional) Dictionary of HTTP Headers to send with the
            :class:`Request`.
        :param cookies: (optional) Dict or CookieJar object to send with the
            :class:`Request`.
        :param files: (optional) Dictionary of ``'filename': file-like-objects``
            for multipart encoding upload.
        :param auth: (optional) Auth tuple or callable to enable
            Basic/Digest/Custom HTTP Auth.
        :param timeout: (optional) How long to wait for the server to send
            data before giving up, as a float, or a :ref:`(connect timeout,
            read timeout) <timeouts>` tuple.
        :param allow_redirects: (optional) Set to True by default.
        :param proxies: (optional) Dictionary mapping protocol or protocol and
            hostname to the URL of the proxy.
        :param hooks: (optional) Dictionary mapping hook name to one event or
            list of events, event must be callable.
        :param stream: (optional) whether to immediately download the response
            content. Defaults to ``False``.
        :param verify: (optional) Either a boolean, in which case it controls whether we verify
            the server's TLS certificate, or a string, in which case it must be a path
            to a CA bundle to use. Defaults to ``True``. When set to
            ``False``, requests will accept any TLS certificate presented by
            the server, and will ignore hostname mismatches and/or expired
            certificates, which will make your application vulnerable to
            man-in-the-middle (MitM) attacks. Setting verify to ``False``
            may be useful during local development or testing.
            It is also possible to put the certificates (directly) in a string or bytes.
        :param cert: (optional) if String, path to ssl client cert file (.pem).
            If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
        """
        ...
    
    def post(
        self,
        url: str,
        *args,
        **kwargs,
    ) -> 'Response':
        return self.request("POST", url, **kwargs)

    @t.overload
    def put(
        self,
        url: str,
        data: 'BodyType' | None = ...,
        *,
        json: t.Any | None = ...,
        params: 'QueryParameterType' | None = ...,
        headers: 'HeadersType' | None = ...,
        cookies: 'CookiesType' | None = ...,
        files: 'MultiPartFilesType' | 'MultiPartFilesAltType' | None = ...,
        auth: 'HttpAuthenticationType' | None = ...,
        timeout: 'TimeoutType' | None = ...,
        allow_redirects: bool = ...,
        proxies: 'ProxyType' | None = ...,
        hooks: 'HookType[PreparedRequest | Response]' | None = ...,
        verify: 'TLSVerifyType' = ...,
        stream: bool = ...,
        cert: 'TLSClientCertType' | None = ...,
    ) -> 'Response':
        r"""Sends a PUT request. Returns :class:`Response` object.

        :param url: URL for the new :class:`Request` object.
        :param params: (optional) Dictionary or bytes to be sent in the query
            string for the :class:`Request`.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param json: (optional) json to send in the body of the
            :class:`Request`.
        :param headers: (optional) Dictionary of HTTP Headers to send with the
            :class:`Request`.
        :param cookies: (optional) Dict or CookieJar object to send with the
            :class:`Request`.
        :param files: (optional) Dictionary of ``'filename': file-like-objects``
            for multipart encoding upload.
        :param auth: (optional) Auth tuple or callable to enable
            Basic/Digest/Custom HTTP Auth.
        :param timeout: (optional) How long to wait for the server to send
            data before giving up, as a float, or a :ref:`(connect timeout,
            read timeout) <timeouts>` tuple.
        :param allow_redirects: (optional) Set to True by default.
        :param proxies: (optional) Dictionary mapping protocol or protocol and
            hostname to the URL of the proxy.
        :param hooks: (optional) Dictionary mapping hook name to one event or
            list of events, event must be callable.
        :param stream: (optional) whether to immediately download the response
            content. Defaults to ``False``.
        :param verify: (optional) Either a boolean, in which case it controls whether we verify
            the server's TLS certificate, or a string, in which case it must be a path
            to a CA bundle to use. Defaults to ``True``. When set to
            ``False``, requests will accept any TLS certificate presented by
            the server, and will ignore hostname mismatches and/or expired
            certificates, which will make your application vulnerable to
            man-in-the-middle (MitM) attacks. Setting verify to ``False``
            may be useful during local development or testing.
            It is also possible to put the certificates (directly) in a string or bytes.
        :param cert: (optional) if String, path to ssl client cert file (.pem).
            If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
        """
        ...
    
    def put(
        self,
        url: str,
        *args,
        **kwargs,
    ) -> 'Response':
        return self.request("PUT", url, **kwargs)

    @t.overload
    def patch(
        self,
        url: str,
        data: 'BodyType' | None = ...,
        *,
        json: t.Any | None = ...,
        params: 'QueryParameterType' | None = ...,
        headers: 'HeadersType' | None = ...,
        cookies: 'CookiesType' | None = ...,
        files: 'MultiPartFilesType' | 'MultiPartFilesAltType' | None = ...,
        auth: 'HttpAuthenticationType' | None = ...,
        timeout: 'TimeoutType' | None = ...,
        allow_redirects: bool = ...,
        proxies: 'ProxyType' | None = ...,
        hooks: 'HookType[PreparedRequest | Response]' | None = ...,
        verify: 'TLSVerifyType' = ...,
        stream: bool = ...,
        cert: 'TLSClientCertType' | None = ...,
    ) -> 'Response':
        r"""Sends a PATCH request. Returns :class:`Response` object.

        :param url: URL for the new :class:`Request` object.
        :param params: (optional) Dictionary or bytes to be sent in the query
            string for the :class:`Request`.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param json: (optional) json to send in the body of the
            :class:`Request`.
        :param headers: (optional) Dictionary of HTTP Headers to send with the
            :class:`Request`.
        :param cookies: (optional) Dict or CookieJar object to send with the
            :class:`Request`.
        :param files: (optional) Dictionary of ``'filename': file-like-objects``
            for multipart encoding upload.
        :param auth: (optional) Auth tuple or callable to enable
            Basic/Digest/Custom HTTP Auth.
        :param timeout: (optional) How long to wait for the server to send
            data before giving up, as a float, or a :ref:`(connect timeout,
            read timeout) <timeouts>` tuple.
        :param allow_redirects: (optional) Set to True by default.
        :param proxies: (optional) Dictionary mapping protocol or protocol and
            hostname to the URL of the proxy.
        :param hooks: (optional) Dictionary mapping hook name to one event or
            list of events, event must be callable.
        :param stream: (optional) whether to immediately download the response
            content. Defaults to ``False``.
        :param verify: (optional) Either a boolean, in which case it controls whether we verify
            the server's TLS certificate, or a string, in which case it must be a path
            to a CA bundle to use. Defaults to ``True``. When set to
            ``False``, requests will accept any TLS certificate presented by
            the server, and will ignore hostname mismatches and/or expired
            certificates, which will make your application vulnerable to
            man-in-the-middle (MitM) attacks. Setting verify to ``False``
            may be useful during local development or testing.
            It is also possible to put the certificates (directly) in a string or bytes.
        :param cert: (optional) if String, path to ssl client cert file (.pem).
            If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
        """
        ...
    
    def patch(
        self,
        url: str,
        *args,
        **kwargs,
    ) -> 'Response':
        return self.request("PATCH", url, **kwargs)

    @t.overload
    def delete(
        self,
        url: str,
        *,
        params: 'QueryParameterType' | None = ...,
        headers: 'HeadersType' | None = ...,
        cookies: 'CookiesType' | None = ...,
        auth: 'HttpAuthenticationType' | None = ...,
        timeout: 'TimeoutType' | None = ...,
        allow_redirects: bool = ...,
        proxies: 'ProxyType' | None = ...,
        hooks: 'HookType[PreparedRequest | Response]' | None = ...,
        verify: 'TLSVerifyType' = ...,
        stream: bool = ...,
        cert: 'TLSClientCertType' | None = ...,
        **kwargs: t.Any,
    ) -> 'Response':
        r"""Sends a DELETE request. Returns :class:`Response` object.

        :param url: URL for the new :class:`Request` object.
        :param params: (optional) Dictionary or bytes to be sent in the query
            string for the :class:`Request`.
        :param headers: (optional) Dictionary of HTTP Headers to send with the
            :class:`Request`.
        :param cookies: (optional) Dict or CookieJar object to send with the
            :class:`Request`.
        :param auth: (optional) Auth tuple or callable to enable
            Basic/Digest/Custom HTTP Auth.
        :param timeout: (optional) How long to wait for the server to send
            data before giving up, as a float, or a :ref:`(connect timeout,
            read timeout) <timeouts>` tuple.
        :param allow_redirects: (optional) Set to True by default.
        :param proxies: (optional) Dictionary mapping protocol or protocol and
            hostname to the URL of the proxy.
        :param hooks: (optional) Dictionary mapping hook name to one event or
            list of events, event must be callable.
        :param stream: (optional) whether to immediately download the response
            content. Defaults to ``False``.
        :param verify: (optional) Either a boolean, in which case it controls whether we verify
            the server's TLS certificate, or a string, in which case it must be a path
            to a CA bundle to use. Defaults to ``True``. When set to
            ``False``, requests will accept any TLS certificate presented by
            the server, and will ignore hostname mismatches and/or expired
            certificates, which will make your application vulnerable to
            man-in-the-middle (MitM) attacks. Setting verify to ``False``
            may be useful during local development or testing.
            It is also possible to put the certificates (directly) in a string or bytes.
        :param cert: (optional) if String, path to ssl client cert file (.pem).
            If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
        """
        ...
    
    def delete(
        self,
        url: str,
        *args,
        **kwargs,
    ) -> 'Response':
        return self.request("DELETE", url, **kwargs)

    
    def gather(self, *responses: 'Response', max_fetch: int | None = None) -> None:
        """
        Call this method to make sure in-flight responses are retrieved efficiently. This is a no-op
        if multiplexed is set to False (which is the default value). Passing a limited set of responses
        will wait for given promises and discard others for later.

        :param max_fetch: Maximal number of response to be fetched before exiting the loop.
               By default, it waits until all pending (lazy) response are resolved.
        """
        return self.io.gather(*responses, max_fetch = max_fetch)
    

    def mount(self, prefix: str, adapter: 'BaseAdapter') -> None:
        """Registers a connection adapter to a prefix.

        Adapters are sorted in descending order by prefix length.
        """
        return self.io.mount(prefix, adapter)
    

    def get_adapter(self, url: str) -> 'BaseAdapter':
        """
        Returns the appropriate connection adapter for the given URL.
        """
        return self.io.get_adapter(url)


    def get_redirect_target(self, resp: 'Response') -> str | None:
        """Receives a Response. Returns a redirect URI or ``None``"""
        return self.io.get_redirect_target(resp)

    """
    Async Client Methods
    """


    @t.overload  # type: ignore[override]
    async def arequest(
        self,
        method: 'HttpMethodType',
        url: str,
        params: 'QueryParameterType' | None = ...,
        data: 'BodyType' | 'AsyncBodyType' | None = ...,
        headers: 'HeadersType' | None = ...,
        cookies: 'CookiesType' | None = ...,
        files: 'MultiPartFilesType' | 'MultiPartFilesAltType' | None = ...,
        auth: 'HttpAuthenticationType' | 'AsyncHttpAuthenticationType' | None = ...,
        timeout: 'TimeoutType' | None = ...,
        allow_redirects: bool = ...,
        proxies: 'ProxyType' | None = ...,
        hooks: 'HookType[PreparedRequest | Response]' | None = ...,
        *,
        stream: t.Literal[True] = True,
        verify: 'TLSVerifyType' | None = ...,
        cert: 'TLSClientCertType' | None = ...,
        json: t.Any | None = ...,
    ) -> 'AsyncResponse': 
        """Constructs a :class:`Request <Request>`, prepares it and sends it.
        Returns :class:`Response <Response>` object.

        :param method: method for the new :class:`Request` object.
        :param url: URL for the new :class:`Request` object.
        :param params: (optional) Dictionary or bytes to be sent in the query
            string for the :class:`Request`.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param json: (optional) json to send in the body of the
            :class:`Request`.
        :param headers: (optional) Dictionary of HTTP Headers to send with the
            :class:`Request`.
        :param cookies: (optional) Dict or CookieJar object to send with the
            :class:`Request`.
        :param files: (optional) Dictionary of ``'filename': file-like-objects``
            for multipart encoding upload.
        :param auth: (optional) Auth tuple or callable to enable
            Basic/Digest/Custom HTTP Auth.
        :param timeout: (optional) How long to wait for the server to send
            data before giving up, as a float, or a :ref:`(connect timeout,
            read timeout) <timeouts>` tuple.
        :param allow_redirects: (optional) Set to True by default.
        :param proxies: (optional) Dictionary mapping protocol or protocol and
            hostname to the URL of the proxy.
        :param hooks: (optional) Dictionary mapping hook name to one event or
            list of events, event must be callable.
        :param stream: (optional) whether to immediately download the response
            content. Defaults to ``False``.
        :param verify: (optional) Either a boolean, in which case it controls whether we verify
            the server's TLS certificate, or a string, in which case it must be a path
            to a CA bundle to use. Defaults to ``True``. When set to
            ``False``, requests will accept any TLS certificate presented by
            the server, and will ignore hostname mismatches and/or expired
            certificates, which will make your application vulnerable to
            man-in-the-middle (MitM) attacks. Setting verify to ``False``
            may be useful during local development or testing.
            It is also possible to put the certificates (directly) in a string or bytes.
        :param cert: (optional) if String, path to ssl client cert file (.pem).
            If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
        
        :return: :class:`AsyncResponse <AsyncResponse>` object
        """
        ...

        
    @t.overload  # type: ignore[override]
    async def arequest(
        self,
        method: 'HttpMethodType',
        url: str,
        params: 'QueryParameterType' | None = ...,
        data: 'BodyType' | 'AsyncBodyType' | None = ...,
        headers: 'HeadersType' | None = ...,
        cookies: 'CookiesType' | None = ...,
        files: 'MultiPartFilesType' | 'MultiPartFilesAltType' | None = ...,
        auth: 'HttpAuthenticationType' | 'AsyncHttpAuthenticationType' | None = ...,
        timeout: 'TimeoutType' | None = ...,
        allow_redirects: bool = ...,
        proxies: 'ProxyType' | None = ...,
        hooks: 'HookType[PreparedRequest | Response]' | None = ...,
        stream: t.Literal[False] = False,
        verify: 'TLSVerifyType' | None = ...,
        cert: 'TLSClientCertType' | None = ...,
        json: t.Any | None = ...,
    ) -> 'Response':
        """Constructs a :class:`Request <Request>`, prepares it and sends it.
        Returns :class:`Response <Response>` object.

        :param method: method for the new :class:`Request` object.
        :param url: URL for the new :class:`Request` object.
        :param params: (optional) Dictionary or bytes to be sent in the query
            string for the :class:`Request`.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param json: (optional) json to send in the body of the
            :class:`Request`.
        :param headers: (optional) Dictionary of HTTP Headers to send with the
            :class:`Request`.
        :param cookies: (optional) Dict or CookieJar object to send with the
            :class:`Request`.
        :param files: (optional) Dictionary of ``'filename': file-like-objects``
            for multipart encoding upload.
        :param auth: (optional) Auth tuple or callable to enable
            Basic/Digest/Custom HTTP Auth.
        :param timeout: (optional) How long to wait for the server to send
            data before giving up, as a float, or a :ref:`(connect timeout,
            read timeout) <timeouts>` tuple.
        :param allow_redirects: (optional) Set to True by default.
        :param proxies: (optional) Dictionary mapping protocol or protocol and
            hostname to the URL of the proxy.
        :param hooks: (optional) Dictionary mapping hook name to one event or
            list of events, event must be callable.
        :param stream: (optional) whether to immediately download the response
            content. Defaults to ``False``.
        :param verify: (optional) Either a boolean, in which case it controls whether we verify
            the server's TLS certificate, or a string, in which case it must be a path
            to a CA bundle to use. Defaults to ``True``. When set to
            ``False``, requests will accept any TLS certificate presented by
            the server, and will ignore hostname mismatches and/or expired
            certificates, which will make your application vulnerable to
            man-in-the-middle (MitM) attacks. Setting verify to ``False``
            may be useful during local development or testing.
            It is also possible to put the certificates (directly) in a string or bytes.
        :param cert: (optional) if String, path to ssl client cert file (.pem).
            If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
        
        :return: :class:`Response <Response>` object
        """
        ...

    async def arequest(  # type: ignore[override]
        self,
        method: 'HttpMethodType',
        url: str,
        **kwargs,
    ) -> 'Response' | 'AsyncResponse':
        url = self._get_url_(url)
        kwargs = self._get_request_kwargs_(method, **kwargs)
        return await self.aio.request(method, url, **kwargs)
    

    @t.overload
    async def aget(
        self,
        url: str,
        *,
        params: 'QueryParameterType' | None = ...,
        headers: 'HeadersType' | None = ...,
        cookies: 'CookiesType' | None = ...,
        auth: 'HttpAuthenticationType' | None = ...,
        timeout: 'TimeoutType' | None = ...,
        allow_redirects: bool = ...,
        proxies: 'ProxyType' | None = ...,
        hooks: 'HookType[PreparedRequest | Response]' | None = ...,
        verify: 'TLSVerifyType' = ...,
        stream: t.Literal[False] = ...,
        cert: 'TLSClientCertType' | None = ...,
        **kwargs: t.Any,
    ) -> 'Response':
        r"""Sends a GET request. Returns :class:`Response` object.

        :param url: URL for the new :class:`Request` object.
        :param params: (optional) Dictionary or bytes to be sent in the query
            string for the :class:`Request`.
        :param headers: (optional) Dictionary of HTTP Headers to send with the
            :class:`Request`.
        :param cookies: (optional) Dict or CookieJar object to send with the
            :class:`Request`.
        :param auth: (optional) Auth tuple or callable to enable
            Basic/Digest/Custom HTTP Auth.
        :param timeout: (optional) How long to wait for the server to send
            data before giving up, as a float, or a :ref:`(connect timeout,
            read timeout) <timeouts>` tuple.
        :param allow_redirects: (optional) Set to True by default.
        :param proxies: (optional) Dictionary mapping protocol or protocol and
            hostname to the URL of the proxy.
        :param hooks: (optional) Dictionary mapping hook name to one event or
            list of events, event must be callable.
        :param stream: (optional) whether to immediately download the response
            content. Defaults to ``False``.
        :param verify: (optional) Either a boolean, in which case it controls whether we verify
            the server's TLS certificate, or a string, in which case it must be a path
            to a CA bundle to use. Defaults to ``True``. When set to
            ``False``, requests will accept any TLS certificate presented by
            the server, and will ignore hostname mismatches and/or expired
            certificates, which will make your application vulnerable to
            man-in-the-middle (MitM) attacks. Setting verify to ``False``
            may be useful during local development or testing.
            It is also possible to put the certificates (directly) in a string or bytes.
        :param cert: (optional) if String, path to ssl client cert file (.pem).
            If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
        """
        ...
    
    @t.overload
    async def aget(
        self,
        url: str,
        *,
        params: 'QueryParameterType' | None = ...,
        headers: 'HeadersType' | None = ...,
        cookies: 'CookiesType' | None = ...,
        auth: 'HttpAuthenticationType' | None = ...,
        timeout: 'TimeoutType' | None = ...,
        allow_redirects: bool = ...,
        proxies: 'ProxyType' | None = ...,
        hooks: 'HookType[PreparedRequest | Response]' | None = ...,
        verify: 'TLSVerifyType' = ...,
        stream: t.Literal[True] = ...,
        cert: 'TLSClientCertType' | None = ...,
        **kwargs: t.Any,
    ) -> 'AsyncResponse':
        r"""Sends a GET request. Returns :class:`Response` object.

        :param url: URL for the new :class:`Request` object.
        :param params: (optional) Dictionary or bytes to be sent in the query
            string for the :class:`Request`.
        :param headers: (optional) Dictionary of HTTP Headers to send with the
            :class:`Request`.
        :param cookies: (optional) Dict or CookieJar object to send with the
            :class:`Request`.
        :param auth: (optional) Auth tuple or callable to enable
            Basic/Digest/Custom HTTP Auth.
        :param timeout: (optional) How long to wait for the server to send
            data before giving up, as a float, or a :ref:`(connect timeout,
            read timeout) <timeouts>` tuple.
        :param allow_redirects: (optional) Set to True by default.
        :param proxies: (optional) Dictionary mapping protocol or protocol and
            hostname to the URL of the proxy.
        :param hooks: (optional) Dictionary mapping hook name to one event or
            list of events, event must be callable.
        :param stream: (optional) whether to immediately download the response
            content. Defaults to ``False``.
        :param verify: (optional) Either a boolean, in which case it controls whether we verify
            the server's TLS certificate, or a string, in which case it must be a path
            to a CA bundle to use. Defaults to ``True``. When set to
            ``False``, requests will accept any TLS certificate presented by
            the server, and will ignore hostname mismatches and/or expired
            certificates, which will make your application vulnerable to
            man-in-the-middle (MitM) attacks. Setting verify to ``False``
            may be useful during local development or testing.
            It is also possible to put the certificates (directly) in a string or bytes.
        :param cert: (optional) if String, path to ssl client cert file (.pem).
            If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
        """
        ...
    
    async def aget(
        self,
        url: str,
        *args,
        **kwargs,
    ) -> 'Response' | 'AsyncResponse':
        return await self.arequest(
            "GET",
            url,
            **kwargs,
        )

    @t.overload
    async def aoptions(
        self,
        url: str,
        *,
        params: 'QueryParameterType' | None = ...,
        headers: 'HeadersType' | None = ...,
        cookies: 'CookiesType' | None = ...,
        auth: 'HttpAuthenticationType' | None = ...,
        timeout: 'TimeoutType' | None = ...,
        allow_redirects: bool = ...,
        proxies: 'ProxyType' | None = ...,
        hooks: 'HookType[PreparedRequest | Response]' | None = ...,
        verify: 'TLSVerifyType' = ...,
        stream: bool = ...,
        cert: 'TLSClientCertType' | None = ...,
        **kwargs: t.Any,
    ) -> 'Response' | 'AsyncResponse':
        r"""Sends a OPTIONS request. Returns :class:`Response` object.

        :param url: URL for the new :class:`Request` object.
        :param params: (optional) Dictionary or bytes to be sent in the query
            string for the :class:`Request`.
        :param headers: (optional) Dictionary of HTTP Headers to send with the
            :class:`Request`.
        :param cookies: (optional) Dict or CookieJar object to send with the
            :class:`Request`.
        :param auth: (optional) Auth tuple or callable to enable
            Basic/Digest/Custom HTTP Auth.
        :param timeout: (optional) How long to wait for the server to send
            data before giving up, as a float, or a :ref:`(connect timeout,
            read timeout) <timeouts>` tuple.
        :param allow_redirects: (optional) Set to True by default.
        :param proxies: (optional) Dictionary mapping protocol or protocol and
            hostname to the URL of the proxy.
        :param hooks: (optional) Dictionary mapping hook name to one event or
            list of events, event must be callable.
        :param stream: (optional) whether to immediately download the response
            content. Defaults to ``False``.
        :param verify: (optional) Either a boolean, in which case it controls whether we verify
            the server's TLS certificate, or a string, in which case it must be a path
            to a CA bundle to use. Defaults to ``True``. When set to
            ``False``, requests will accept any TLS certificate presented by
            the server, and will ignore hostname mismatches and/or expired
            certificates, which will make your application vulnerable to
            man-in-the-middle (MitM) attacks. Setting verify to ``False``
            may be useful during local development or testing.
            It is also possible to put the certificates (directly) in a string or bytes.
        :param cert: (optional) if String, path to ssl client cert file (.pem).
            If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
        """
        ...
    
    async def aoptions(
        self,
        url: str,
        *args,
        **kwargs,
    ) -> 'Response' | 'AsyncResponse':
        return await self.arequest(
            "OPTIONS",
            url,
            **kwargs,
        )

    @t.overload
    async def ahead(
        self,
        url: str,
        *,
        params: 'QueryParameterType' | None = ...,
        headers: 'HeadersType' | None = ...,
        cookies: 'CookiesType' | None = ...,
        auth: 'HttpAuthenticationType' | None = ...,
        timeout: 'TimeoutType' | None = ...,
        allow_redirects: bool = ...,
        proxies: 'ProxyType' | None = ...,
        hooks: 'HookType[PreparedRequest | Response]' | None = ...,
        verify: 'TLSVerifyType' = ...,
        stream: bool = ...,
        cert: 'TLSClientCertType' | None = ...,
        **kwargs: t.Any,
    ) -> 'Response' | 'AsyncResponse':
        r"""Sends a HEAD request. Returns :class:`Response` object.

        :param url: URL for the new :class:`Request` object.
        :param params: (optional) Dictionary or bytes to be sent in the query
            string for the :class:`Request`.
        :param headers: (optional) Dictionary of HTTP Headers to send with the
            :class:`Request`.
        :param cookies: (optional) Dict or CookieJar object to send with the
            :class:`Request`.
        :param auth: (optional) Auth tuple or callable to enable
            Basic/Digest/Custom HTTP Auth.
        :param timeout: (optional) How long to wait for the server to send
            data before giving up, as a float, or a :ref:`(connect timeout,
            read timeout) <timeouts>` tuple.
        :param allow_redirects: (optional) Set to False by default.
        :param proxies: (optional) Dictionary mapping protocol or protocol and
            hostname to the URL of the proxy.
        :param hooks: (optional) Dictionary mapping hook name to one event or
            list of events, event must be callable.
        :param stream: (optional) whether to immediately download the response
            content. Defaults to ``False``.
        :param verify: (optional) Either a boolean, in which case it controls whether we verify
            the server's TLS certificate, or a string, in which case it must be a path
            to a CA bundle to use. Defaults to ``True``. When set to
            ``False``, requests will accept any TLS certificate presented by
            the server, and will ignore hostname mismatches and/or expired
            certificates, which will make your application vulnerable to
            man-in-the-middle (MitM) attacks. Setting verify to ``False``
            may be useful during local development or testing.
            It is also possible to put the certificates (directly) in a string or bytes.
        :param cert: (optional) if String, path to ssl client cert file (.pem).
            If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
        """
        ...
    
    async def ahead(
        self,
        url: str,
        *args,
        **kwargs,
    ) -> 'Response' | 'AsyncResponse':
        return await self.arequest(
            "HEAD",
            url,
            **kwargs,
        )

    @t.overload
    async def apost(
        self,
        url: str,
        data: 'BodyType' | None = ...,
        json: t.Any | None = ...,
        *,
        params: 'QueryParameterType' | None = ...,
        headers: 'HeadersType' | None = ...,
        cookies: 'CookiesType' | None = ...,
        files: 'MultiPartFilesType' | 'MultiPartFilesAltType' | None = ...,
        auth: 'HttpAuthenticationType' | None = ...,
        timeout: 'TimeoutType' | None = ...,
        allow_redirects: bool = ...,
        proxies: 'ProxyType' | None = ...,
        hooks: 'HookType[PreparedRequest | Response]' | None = ...,
        verify: 'TLSVerifyType' = ...,
        stream: t.Literal[False] = ...,
        cert: 'TLSClientCertType' | None = ...,
    ) -> 'Response':
        r"""Sends a POST request. Returns :class:`Response` object.

        :param url: URL for the new :class:`Request` object.
        :param params: (optional) Dictionary or bytes to be sent in the query
            string for the :class:`Request`.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param json: (optional) json to send in the body of the
            :class:`Request`.
        :param headers: (optional) Dictionary of HTTP Headers to send with the
            :class:`Request`.
        :param cookies: (optional) Dict or CookieJar object to send with the
            :class:`Request`.
        :param files: (optional) Dictionary of ``'filename': file-like-objects``
            for multipart encoding upload.
        :param auth: (optional) Auth tuple or callable to enable
            Basic/Digest/Custom HTTP Auth.
        :param timeout: (optional) How long to wait for the server to send
            data before giving up, as a float, or a :ref:`(connect timeout,
            read timeout) <timeouts>` tuple.
        :param allow_redirects: (optional) Set to True by default.
        :param proxies: (optional) Dictionary mapping protocol or protocol and
            hostname to the URL of the proxy.
        :param hooks: (optional) Dictionary mapping hook name to one event or
            list of events, event must be callable.
        :param stream: (optional) whether to immediately download the response
            content. Defaults to ``False``.
        :param verify: (optional) Either a boolean, in which case it controls whether we verify
            the server's TLS certificate, or a string, in which case it must be a path
            to a CA bundle to use. Defaults to ``True``. When set to
            ``False``, requests will accept any TLS certificate presented by
            the server, and will ignore hostname mismatches and/or expired
            certificates, which will make your application vulnerable to
            man-in-the-middle (MitM) attacks. Setting verify to ``False``
            may be useful during local development or testing.
            It is also possible to put the certificates (directly) in a string or bytes.
        :param cert: (optional) if String, path to ssl client cert file (.pem).
            If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
        """
        ...

    @t.overload
    async def apost(
        self,
        url: str,
        data: 'BodyType' | None = ...,
        json: t.Any | None = ...,
        *,
        params: 'QueryParameterType' | None = ...,
        headers: 'HeadersType' | None = ...,
        cookies: 'CookiesType' | None = ...,
        files: 'MultiPartFilesType' | 'MultiPartFilesAltType' | None = ...,
        auth: 'HttpAuthenticationType' | None = ...,
        timeout: 'TimeoutType' | None = ...,
        allow_redirects: bool = ...,
        proxies: 'ProxyType' | None = ...,
        hooks: 'HookType[PreparedRequest | Response]' | None = ...,
        verify: 'TLSVerifyType' = ...,
        stream: t.Literal[True] = ...,
        cert: 'TLSClientCertType' | None = ...,
    ) -> 'AsyncResponse':
        r"""Sends a POST request. Returns :class:`Response` object.

        :param url: URL for the new :class:`Request` object.
        :param params: (optional) Dictionary or bytes to be sent in the query
            string for the :class:`Request`.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param json: (optional) json to send in the body of the
            :class:`Request`.
        :param headers: (optional) Dictionary of HTTP Headers to send with the
            :class:`Request`.
        :param cookies: (optional) Dict or CookieJar object to send with the
            :class:`Request`.
        :param files: (optional) Dictionary of ``'filename': file-like-objects``
            for multipart encoding upload.
        :param auth: (optional) Auth tuple or callable to enable
            Basic/Digest/Custom HTTP Auth.
        :param timeout: (optional) How long to wait for the server to send
            data before giving up, as a float, or a :ref:`(connect timeout,
            read timeout) <timeouts>` tuple.
        :param allow_redirects: (optional) Set to True by default.
        :param proxies: (optional) Dictionary mapping protocol or protocol and
            hostname to the URL of the proxy.
        :param hooks: (optional) Dictionary mapping hook name to one event or
            list of events, event must be callable.
        :param stream: (optional) whether to immediately download the response
            content. Defaults to ``False``.
        :param verify: (optional) Either a boolean, in which case it controls whether we verify
            the server's TLS certificate, or a string, in which case it must be a path
            to a CA bundle to use. Defaults to ``True``. When set to
            ``False``, requests will accept any TLS certificate presented by
            the server, and will ignore hostname mismatches and/or expired
            certificates, which will make your application vulnerable to
            man-in-the-middle (MitM) attacks. Setting verify to ``False``
            may be useful during local development or testing.
            It is also possible to put the certificates (directly) in a string or bytes.
        :param cert: (optional) if String, path to ssl client cert file (.pem).
            If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
        """
        ...
    
    async def apost(
        self,
        url: str,
        *args,
        **kwargs,
    ) -> 'Response' | 'AsyncResponse':
        return await self.arequest("POST", url, **kwargs)

    @t.overload
    async def aput(
        self,
        url: str,
        data: 'BodyType' | None = ...,
        *,
        json: t.Any | None = ...,
        params: 'QueryParameterType' | None = ...,
        headers: 'HeadersType' | None = ...,
        cookies: 'CookiesType' | None = ...,
        files: 'MultiPartFilesType' | 'MultiPartFilesAltType' | None = ...,
        auth: 'HttpAuthenticationType' | None = ...,
        timeout: 'TimeoutType' | None = ...,
        allow_redirects: bool = ...,
        proxies: 'ProxyType' | None = ...,
        hooks: 'HookType[PreparedRequest | Response]' | None = ...,
        verify: 'TLSVerifyType' = ...,
        stream: bool = ...,
        cert: 'TLSClientCertType' | None = ...,
    ) -> 'Response' | 'AsyncResponse':
        r"""Sends a PUT request. Returns :class:`Response` object.

        :param url: URL for the new :class:`Request` object.
        :param params: (optional) Dictionary or bytes to be sent in the query
            string for the :class:`Request`.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param json: (optional) json to send in the body of the
            :class:`Request`.
        :param headers: (optional) Dictionary of HTTP Headers to send with the
            :class:`Request`.
        :param cookies: (optional) Dict or CookieJar object to send with the
            :class:`Request`.
        :param files: (optional) Dictionary of ``'filename': file-like-objects``
            for multipart encoding upload.
        :param auth: (optional) Auth tuple or callable to enable
            Basic/Digest/Custom HTTP Auth.
        :param timeout: (optional) How long to wait for the server to send
            data before giving up, as a float, or a :ref:`(connect timeout,
            read timeout) <timeouts>` tuple.
        :param allow_redirects: (optional) Set to True by default.
        :param proxies: (optional) Dictionary mapping protocol or protocol and
            hostname to the URL of the proxy.
        :param hooks: (optional) Dictionary mapping hook name to one event or
            list of events, event must be callable.
        :param stream: (optional) whether to immediately download the response
            content. Defaults to ``False``.
        :param verify: (optional) Either a boolean, in which case it controls whether we verify
            the server's TLS certificate, or a string, in which case it must be a path
            to a CA bundle to use. Defaults to ``True``. When set to
            ``False``, requests will accept any TLS certificate presented by
            the server, and will ignore hostname mismatches and/or expired
            certificates, which will make your application vulnerable to
            man-in-the-middle (MitM) attacks. Setting verify to ``False``
            may be useful during local development or testing.
            It is also possible to put the certificates (directly) in a string or bytes.
        :param cert: (optional) if String, path to ssl client cert file (.pem).
            If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
        """
        ...
    
    async def aput(
        self,
        url: str,
        *args,
        **kwargs,
    ) -> 'Response' | 'AsyncResponse':
        return await self.arequest("PUT", url, **kwargs)

    @t.overload
    async def apatch(
        self,
        url: str,
        data: 'BodyType' | None = ...,
        *,
        json: t.Any | None = ...,
        params: 'QueryParameterType' | None = ...,
        headers: 'HeadersType' | None = ...,
        cookies: 'CookiesType' | None = ...,
        files: 'MultiPartFilesType' | 'MultiPartFilesAltType' | None = ...,
        auth: 'HttpAuthenticationType' | None = ...,
        timeout: 'TimeoutType' | None = ...,
        allow_redirects: bool = ...,
        proxies: 'ProxyType' | None = ...,
        hooks: 'HookType[PreparedRequest | Response]' | None = ...,
        verify: 'TLSVerifyType' = ...,
        stream: bool = ...,
        cert: 'TLSClientCertType' | None = ...,
    ) -> 'Response' | 'AsyncResponse':
        r"""Sends a PATCH request. Returns :class:`Response` object.

        :param url: URL for the new :class:`Request` object.
        :param params: (optional) Dictionary or bytes to be sent in the query
            string for the :class:`Request`.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param json: (optional) json to send in the body of the
            :class:`Request`.
        :param headers: (optional) Dictionary of HTTP Headers to send with the
            :class:`Request`.
        :param cookies: (optional) Dict or CookieJar object to send with the
            :class:`Request`.
        :param files: (optional) Dictionary of ``'filename': file-like-objects``
            for multipart encoding upload.
        :param auth: (optional) Auth tuple or callable to enable
            Basic/Digest/Custom HTTP Auth.
        :param timeout: (optional) How long to wait for the server to send
            data before giving up, as a float, or a :ref:`(connect timeout,
            read timeout) <timeouts>` tuple.
        :param allow_redirects: (optional) Set to True by default.
        :param proxies: (optional) Dictionary mapping protocol or protocol and
            hostname to the URL of the proxy.
        :param hooks: (optional) Dictionary mapping hook name to one event or
            list of events, event must be callable.
        :param stream: (optional) whether to immediately download the response
            content. Defaults to ``False``.
        :param verify: (optional) Either a boolean, in which case it controls whether we verify
            the server's TLS certificate, or a string, in which case it must be a path
            to a CA bundle to use. Defaults to ``True``. When set to
            ``False``, requests will accept any TLS certificate presented by
            the server, and will ignore hostname mismatches and/or expired
            certificates, which will make your application vulnerable to
            man-in-the-middle (MitM) attacks. Setting verify to ``False``
            may be useful during local development or testing.
            It is also possible to put the certificates (directly) in a string or bytes.
        :param cert: (optional) if String, path to ssl client cert file (.pem).
            If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
        """
        ...
    
    async def apatch(
        self,
        url: str,
        *args,
        **kwargs,
    ) -> 'Response' | 'AsyncResponse':
        return await self.arequest("PATCH", url, **kwargs)


    @t.overload
    async def adelete(
        self,
        url: str,
        *,
        params: 'QueryParameterType' | None = ...,
        headers: 'HeadersType' | None = ...,
        cookies: 'CookiesType' | None = ...,
        auth: 'HttpAuthenticationType' | None = ...,
        timeout: 'TimeoutType' | None = ...,
        allow_redirects: bool = ...,
        proxies: 'ProxyType' | None = ...,
        hooks: 'HookType[PreparedRequest | Response]' | None = ...,
        verify: 'TLSVerifyType' = ...,
        stream: bool = ...,
        cert: 'TLSClientCertType' | None = ...,
        **kwargs: t.Any,
    ) -> 'Response' | 'AsyncResponse':
        r"""Sends a DELETE request. Returns :class:`Response` object.

        :param url: URL for the new :class:`Request` object.
        :param params: (optional) Dictionary or bytes to be sent in the query
            string for the :class:`Request`.
        :param headers: (optional) Dictionary of HTTP Headers to send with the
            :class:`Request`.
        :param cookies: (optional) Dict or CookieJar object to send with the
            :class:`Request`.
        :param auth: (optional) Auth tuple or callable to enable
            Basic/Digest/Custom HTTP Auth.
        :param timeout: (optional) How long to wait for the server to send
            data before giving up, as a float, or a :ref:`(connect timeout,
            read timeout) <timeouts>` tuple.
        :param allow_redirects: (optional) Set to True by default.
        :param proxies: (optional) Dictionary mapping protocol or protocol and
            hostname to the URL of the proxy.
        :param hooks: (optional) Dictionary mapping hook name to one event or
            list of events, event must be callable.
        :param stream: (optional) whether to immediately download the response
            content. Defaults to ``False``.
        :param verify: (optional) Either a boolean, in which case it controls whether we verify
            the server's TLS certificate, or a string, in which case it must be a path
            to a CA bundle to use. Defaults to ``True``. When set to
            ``False``, requests will accept any TLS certificate presented by
            the server, and will ignore hostname mismatches and/or expired
            certificates, which will make your application vulnerable to
            man-in-the-middle (MitM) attacks. Setting verify to ``False``
            may be useful during local development or testing.
            It is also possible to put the certificates (directly) in a string or bytes.
        :param cert: (optional) if String, path to ssl client cert file (.pem).
            If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
        """
        ...
    
    async def adelete(
        self,
        url: str,
        *args,
        **kwargs,
    ) -> 'Response' | 'AsyncResponse':
        return await self.arequest("DELETE", url, **kwargs)

    async def agather(
        self, 
        *responses: 'Response', 
        max_fetch: int | None = None
    ) -> None:  # type: ignore[override]
        """
        Call this method to make sure in-flight responses are retrieved efficiently. This is a no-op
        if multiplexed is set to False (which is the default value). Passing a limited set of responses
        will wait for given promises and discard others for later.

        :param max_fetch: Maximal number of response to be fetched before exiting the loop.
               By default, it waits until all pending (lazy) response are resolved.
        """
        return await self.aio.gather(*responses, max_fetch = max_fetch)


    def amount(self, prefix: str, adapter: 'AsyncBaseAdapter') -> None:  # type: ignore[override]
        """Registers a connection adapter to a prefix.

        Adapters are sorted in descending order by prefix length.
        """
        return self.aio.mount(prefix, adapter)  # type: ignore[arg-type]
    
    def aget_adapter(self, url: str) -> 'AsyncBaseAdapter':  # type: ignore[override]
        """
        Returns the appropriate connection adapter for the given URL.
        """
        return self.aio.get_adapter(url)  # type: ignore[arg-type]
    

    """
    Client Class Methods
    """

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc, value, tb):
        await self.aclose()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()


from __future__ import annotations

"""Public façade for the LazyOps dependency resolution helpers."""

from .base import (
    LazyLib,
    LazyLibType,
    PkgInstall,
    require_missing_wrapper,
    resolve_missing,
    resolve_missing_custom,
)

__all__ = [
    "PkgInstall",
    "LazyLibType",
    "LazyLib",
    "resolve_missing",
    "resolve_missing_custom",
    "require_missing_wrapper",
]



from __future__ import annotations

"""Shared mixins for auto-registering LazyOps components."""

import abc
import copy
import typing as t

if t.TYPE_CHECKING:
    from pathlib import Path
    from pydantic import ConfigDict

__all__ = [
    'ExtraMeta',
    'RegisteredClient',
    'RegisteredObject',
    'RegisteredSettings',
    'add_extra_meta',
]


def add_extra_meta(obj: t.Type['ExtraMeta']) -> None:
    """Populate bookkeeping metadata on registration-aware classes.

    Args:
        obj: Class that derives from :class:`ExtraMeta`.
    """

    import inspect
    from pathlib import Path

    from lzl.io.ser import get_object_classname
    from lzl.logging import logger

    cls_name = get_object_classname(obj, is_type=True)
    cls_module = obj.__module__.split('.')[0]
    module_lib_path = Path(inspect.getfile(obj)).parent
    obj._rxtra['module'] = cls_module
    obj._rxtra['cls_name'] = cls_name
    obj._rxtra['module_lib_path'] = module_lib_path
    logger.debug('Registered meta for %s.%s', cls_module, cls_name)
    if '__main__' in cls_module:
        return

    p = module_lib_path
    m_path, iters = None, 0
    while p.name != cls_module and iters < 4:
        p = p.parent
        iters += 1
        if p.name == cls_module:
            m_path = p
            break
    if m_path is not None:
        obj._rxtra['module_path'] = m_path


class ExtraMeta(abc.ABC):
    """Base class that captures module and path metadata for registries."""

    _rxtra: t.Dict[str, t.Any] = {}

    def __init_subclass__(cls, **kwargs: t.Any) -> None:  # pragma: no cover - instrumentation
        add_extra_meta(cls)
        super().__init_subclass__(**kwargs)

    @property
    def module_path(self) -> 'Path':
        """Absolute path to the package containing the class."""

        return self._rxtra.get('module_path')

    @property
    def module_lib_path(self) -> 'Path':
        """Filesystem path of the module where the class is defined."""

        return self._rxtra.get('module_lib_path')

    @property
    def module_name(self) -> str:
        """Dotted module name derived from the class namespace."""

        return self._rxtra.get('module')

    @property
    def cls_name(self) -> str:
        """Original class name captured during registration."""

        return self._rxtra.get('cls_name')


class RegisteredClient(abc.ABC):
    """Mixin that auto-registers client classes with :mod:`lzo.registry`."""

    name: t.Optional[str] = None
    _rxtra: t.Dict[str, t.Any] = {}
    _rmodule: t.Optional[str] = None
    _rsubmodule: t.Optional[str] = None

    if t.TYPE_CHECKING:
        enable_registration: bool

    def __init_subclass__(cls, **kwargs: t.Any) -> None:  # pragma: no cover - registration glue
        if not hasattr(cls, 'enable_registration') or cls.enable_registration is True:
            from lzo.registry.clients import register_client

            register_client(cls)
        super().__init_subclass__(**kwargs)

    @classmethod
    def configure_registered(cls, **kwargs: t.Any) -> 'RegisteredClient':
        """Return a copy of ``cls`` with overridden registration metadata."""

        new_cls = copy.deepcopy(cls)
        if kwargs.get('module'):
            new_cls._rmodule = kwargs.pop('module')
        if kwargs.get('submodule'):
            new_cls._rsubmodule = kwargs.pop('submodule')
        if kwargs.get('name'):
            new_cls.name = kwargs.pop('name')
        if kwargs:
            new_cls._rxtra.update(kwargs)
        return new_cls


class RegisteredObject(abc.ABC):
    """Mixin that registers singleton-like objects for lazy access."""

    name: t.Optional[str] = None
    kind: t.Optional[str] = None

    _rxtra: t.Dict[str, t.Any] = {}
    _rmodule: t.Optional[str] = None

    def __init_subclass__(cls, **kwargs: t.Any) -> None:  # pragma: no cover - registration glue
        from lzo.registry.clients import register_client

        register_client(cls)
        super().__init_subclass__(**kwargs)

    @classmethod
    def configure_registered(cls, **kwargs: t.Any) -> 'RegisteredObject':
        """Return a copy of ``cls`` with custom registration metadata."""

        new_cls = copy.deepcopy(cls)
        if kwargs.get('module'):
            new_cls._rmodule = kwargs.pop('module')
        if kwargs.get('kind'):
            new_cls.kind = kwargs.pop('kind')
        if kwargs.get('name'):
            new_cls.name = kwargs.pop('name')
        if kwargs:
            new_cls._rxtra.update(kwargs)
        return new_cls


class RegisteredSettings(abc.ABC):
    """Mixin that registers pydantic settings classes with the registry."""

    _rmodule: t.Optional[str] = None
    _rsubmodule: t.Optional[str] = None
    _rxtra: t.Dict[str, t.Any] = {}

    def __init_subclass__(cls, **kwargs: 'ConfigDict') -> None:  # pragma: no cover - registration glue
        from lzo.registry.settings import register_settings

        register_settings(cls)
        super().__init_subclass__(**kwargs)

    @classmethod
    def configure_registered(cls, **kwargs: t.Any) -> 'RegisteredSettings':
        """Return a copy of ``cls`` with overridden registry metadata."""

        new_cls = copy.deepcopy(cls)
        if kwargs.get('module'):
            new_cls._rmodule = kwargs.pop('module')
        if kwargs.get('submodule'):
            new_cls._rsubmodule = kwargs.pop('submodule')
        if kwargs:
            new_cls._rxtra.update(kwargs)
        return new_cls

    @property
    def _rverbose(self) -> bool:
        """Flag controlling whether verbose logging is enabled for this class."""

        return self._rxtra.get('verbose', False)

    if t.TYPE_CHECKING:

        @property
        def module_path(self) -> Path:  # pragma: no cover - typing helper
            ...

        @property
        def module_config_path(self) -> Path:  # pragma: no cover - typing helper
            ...

        @property
        def module_name(self) -> str:  # pragma: no cover - typing helper
            ...



from __future__ import annotations

"""
Host Resources
"""

import typing as t
from lzl import load
from .utils import parse_memory_metric_to_bs, safe_float

if t.TYPE_CHECKING:
    import psutil
    from pydantic.types import ByteSize
else:
    psutil = load.lazy_load("psutil")


_resource_dict_mapping = {
    'cpu_count': int,
    
    'memory_total': parse_memory_metric_to_bs,
    'memory_used': parse_memory_metric_to_bs,
    'memory_free': parse_memory_metric_to_bs,

    'utilization_cpu': safe_float,
    'utilization_memory': safe_float,
}

ResourceData = t.Dict[str, t.Union[str, int, float, 'ByteSize']]

def get_resource_data() -> ResourceData:
    """
    Returns the resource data of the host system.
    """
    return {
        'cpu_count': psutil.cpu_count(),
        'memory_total': parse_memory_metric_to_bs(psutil.virtual_memory().total),
        'memory_used': parse_memory_metric_to_bs(psutil.virtual_memory().used),
        'memory_free': parse_memory_metric_to_bs(psutil.virtual_memory().free),
        'utilization_cpu': safe_float(psutil.cpu_percent()),
        'utilization_memory': safe_float(psutil.virtual_memory().percent),
    }


from __future__ import annotations

"""
Kinde OAuth2 Config
"""

from ...types.config import BaseOAuth2Config, eproperty
from typing import Any, Dict, Optional, Union, Tuple, TYPE_CHECKING, List, Type, Annotated
from pydantic import Field
from lzo.utils import create_unique_id
from .token import AccessToken, IDToken, OAuth2Token

_default_scopes = [
    'openid',
    'profile',
    'email',
    'offline',
]

class KindeOAuth2Config(BaseOAuth2Config):
    """
    Kinde Auth Settings
    """
    name: Optional[str] = 'kinde'
    scopes: Optional[List[str]] = Field(default = _default_scopes)

    class Config(BaseOAuth2Config.Config):
        env_prefix = "KINDE_"

    @eproperty
    def m2m_audience(self) -> str:
        """
        Returns the m2m audience
        """
        return f'{self.issuer_url}/api'
    

    @classmethod
    def get_default_token_cls(cls, name: str) -> Union[Type[AccessToken], Type[IDToken], Type[OAuth2Token]]:
        """
        Returns the default token class
        """
        if name == 'access_token':
            return AccessToken
        return IDToken if name == 'id_token' else OAuth2Token


    def get_auth_params(
        self,
        redirect_url: str,
        scope: Optional[str] = None,
        state: Optional[str] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Returns the auth params
        """
        scope = scope or self.get_scopes()
        state = state or create_unique_id(alph_only=True)
        return {
            'response_type': 'code',
            'client_id': self.client_id,
            'redirect_uri': redirect_url,
            'scope': scope,
            'state': state,
            'code_challenge': self.create_code_challenge(self.code_verifier),
            'code_challenge_method': self.code_challenge_method,
        }
    
    def get_fetch_token_params(
        self,
        code: str,
        redirect_url: str,
        state: Optional[str] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Returns the fetch token params
        """
        return {
            'code': code,
            'code_verifier': self.code_verifier,
            'state': state,
            'redirect_uri': redirect_url,
            'grant_type': 'authorization_code',
        }
    
    def get_refresh_token_params(
        self,
        refresh_token: str,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Returns the refresh token params
        """
        return {
            'refresh_token': refresh_token,
            'client_id': self.client_id,
            'client_secret': self.client_secret,
            'grant_type': 'refresh_token',
        }

    def get_headers(
        self,
        endpoint_type: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Returns the headers
        """
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
            'Accept': 'application/json',
        }
        if endpoint_type == 'token':
            headers['Authorization'] = f'Basic {self.encode_client_secret_basic(self.client_id, self.client_secret)}'
        return headers




from __future__ import annotations

"""
Base Persistence Backend
"""
import contextlib
import collections.abc
from threading import Lock
from asyncio import Lock as AsyncLock
from pathlib import Path
from pydantic import BaseModel
from typing import TypeVar, Generic, Any, Dict, Optional, Union, Iterable, List, Type, ItemsView, TYPE_CHECKING
from lzo.utils import create_unique_id, logger
from lzl.pool import ThreadPool
from lzo.utils.hashing import create_object_hash, create_hash_from_args_and_kwargs
from lzl.io.ser import SerT, get_serializer, SchemaType

if TYPE_CHECKING:
    from lzo.types import BaseSettings
    from lzl.io.ser import ObjectValue



class BaseStatefulBackend(collections.abc.MutableMapping):
    """
    Base Stateful Backend
    """
    serializer: Optional[SerT] = None
    serializer_type: Optional[str] = None
    serializer_kwargs: Optional[Dict[str, Any]] = {}
    _extra: Dict[str, Any] = {}

    def __init__(
        self,
        serializer: Optional[str] = None,
        serializer_kwargs: Optional[Dict[str, Any]] = None,
        base_key: Optional[str] = None,
        async_enabled: Optional[bool] = False,
        settings: Optional['BaseSettings'] = None,
        **kwargs,
    ):
        """
        Initializes the backend
        """
        self.base_key = base_key
        self.async_enabled = async_enabled
        self.settings = settings
        if serializer is not None:
            self.serializer_type = serializer
        if serializer_kwargs is not None:
            self.serializer_kwargs = serializer_kwargs
        if self.serializer is None:
            self.serializer = get_serializer(self.serializer_type, **self.serializer_kwargs)
        self._lock: Optional[Lock] = None
        self._alock: Optional[AsyncLock] = None
        self._kwargs = kwargs


    def get_key(self, key: str) -> str:
        """
        Gets a Key
        """
        return f'{self.base_key}.{key}' if self.base_key else key

    def encode_value(self, value: Union[Any, SchemaType], _raw: Optional[bool] = None, **kwargs) -> Union[str, bytes]:
        """
        Encodes a Value
        """
        return value if _raw else self.serializer.encode(value, **kwargs)
    
    async def aencode_value(self, value: Union[Any, SchemaType], _raw: Optional[bool] = None,  **kwargs) -> Union[str, bytes]:
        """
        Encodes a Value
        """
        return value if _raw else await self.serializer.aencode(value, **kwargs)

    def decode_value(self, value: Union[str, bytes], _raw: Optional[bool] = None, **kwargs) -> Any:
        """
        Decodes a Value
        """
        return value if _raw else self.serializer.decode(value, **kwargs)
    
    async def adecode_value(self, value: Union[str, bytes], _raw: Optional[bool] = None, **kwargs) -> Any:
        """
        Decodes a Value
        """
        return value if _raw else await self.serializer.adecode(value, **kwargs)
    
    def create_hash(self, obj: 'ObjectValue') -> str:
        """
        Creates a Hash
        """
        return self.serializer.create_hash(obj) if self.serializer is not None else \
            create_object_hash(obj)

    
    def _precheck(self, **kwargs):
        """
        Run a precheck operation
        """
        pass

    def get(self, key: str, default: Optional[Any] = None, _raw: Optional[bool] = None, **kwargs) -> Optional[Any]:
        """
        Gets a Value from the DB
        """
        raise NotImplementedError
            
    def get_values(self, keys: Iterable[str]) -> List[Any]:
        """
        Gets a Value from the DB
        """
        raise NotImplementedError


    def set(self, key: str, value: Any, ex: Optional[int] = None, _raw: Optional[bool] = None, **kwargs) -> None:
        """
        Saves a Value to the DB
        """
        raise NotImplementedError


    def set_batch(self, data: Dict[str, Any], ex: Optional[int] = None, _raw: Optional[bool] = None, **kwargs) -> None:
        """
        Saves a Value to the DB
        """
        raise NotImplementedError


    def delete(self, key: str, **kwargs) -> None:
        """
        Deletes a Value from the DB
        """
        raise NotImplementedError


    def clear(self, *keys: str, **kwargs):
        """
        Clears the Cache
        """
        raise NotImplementedError

    
    async def aget(self, key: str, default: Optional[Any] = None, _raw: Optional[bool] = None, **kwargs) -> Optional[Any]:
        """
        Gets a Value from the DB
        """
        return await ThreadPool.run_async(self.get, key, default, _raw = _raw, **kwargs)

    async def aget_values(self, keys: Iterable[str], **kwargs) -> List[Any]:
        """
        Gets a Value from the DB
        """
        return await ThreadPool.run_async(self.get_values, keys, **kwargs)

    
    async def aset(self, key: str, value: Any, ex: Optional[int] = None, _raw: Optional[bool] = None, **kwargs) -> None:
        """
        Saves a Value to the DB
        """
        return await ThreadPool.run_async(self.set, key, value, ex, _raw = _raw, **kwargs)


    async def aset_batch(self, data: Dict[str, Any], ex: Optional[int] = None, **kwargs) -> None:
        """
        Saves a Value to the DB
        """
        return await ThreadPool.run_async(self.set_batch, data, ex, **kwargs)

    async def adelete(self, key: str, **kwargs) -> None:
        """
        Deletes a Value from the DB
        """
        return await ThreadPool.run_async(self.delete, key, **kwargs)

    async def aclear(self, *keys: str, **kwargs):
        """
        Clears the Cache
        """
        return await ThreadPool.run_async(self.clear, *keys, **kwargs)

    def iterate(self, **kwargs) -> Iterable[Any]:
        """
        Iterates over the Cache
        """
        raise NotImplementedError


    def get_all_data(self, exclude_base_key: Optional[bool] = False, **kwargs) -> Dict[str, Any]:
        """
        Loads all the Data
        """
        raise NotImplementedError


    def get_all_keys(self, exclude_base_key: Optional[bool] = False, **kwargs) -> List[str]:
        """
        Returns all the Keys
        """
        raise NotImplementedError


    def get_all_values(self, **kwargs) -> List[Any]:
        """
        Returns all the Values
        """
        raise NotImplementedError


    async def aget_all_data(self, exclude_base_key: Optional[bool] = False, **kwargs) -> Dict[str, Any]:
        """
        Loads all the Data
        """
        return await ThreadPool.run_async(self.get_all_data, exclude_base_key = exclude_base_key, **kwargs)

    async def aget_all_keys(self, exclude_base_key: Optional[bool] = False, **kwargs) -> List[str]:
        """
        Returns all the Keys
        """
        return await ThreadPool.run_async(self.get_all_keys, exclude_base_key = exclude_base_key, **kwargs)

    async def aget_all_values(self,  **kwargs) -> List[Any]:
        """
        Returns all the Values
        """
        return await ThreadPool.run_async(self.get_all_values, **kwargs)


    def pop(self, key, default = None, **kwargs):
        """
        Pops an Item from the Cache
        """
        self._precheck()
        value = self.get(key, **kwargs)
        if self.async_enabled:
            ThreadPool.create_background_task(self.adelete(key))
        else:
            self.delete(key)
        return value if value is not None else default
    
    async def apop(self, key, default = None, **kwargs):
        """
        Pops an Item from the Cache
        """
        self._precheck()
        value = await self.aget(key, **kwargs)
        await self.adelete(key)
        return value if value is not None else default

    def contains(self, key: str, **kwargs) -> bool:
        """
        Returns True if the Cache contains the Key
        """
        raise NotImplementedError

    async def acontains(self, key: str, **kwargs) -> bool:
        """
        Returns True if the Cache contains the Key
        """
        return await ThreadPool.run_async(self.contains, key, **kwargs)

    def keys(self, **kwargs) -> Iterable[str]:
        """
        Returns the Keys
        """
        return iter(self.get_all_keys(True, **kwargs))

    def get_keys(self, pattern: str, exclude_base_key: Optional[bool] = None, **kwargs) -> List[str]:
        """
        Returns the Keys
        """
        raise NotImplementedError
    
    def values(self, **kwargs) -> Iterable[Any]:
        """
        Returns the Values
        """
        return iter(self.get_all_values(**kwargs))
    
    def items(self, iterable: Optional[bool] = True, **kwargs) -> Union[Iterable[ItemsView], Dict[str, Any]]:
        """
        Returns the Items
        """
        items = self.get_all_data(True, **kwargs)
        return items.items() if iterable else items
    
    async def akeys(self, **kwargs) -> Iterable[Any]:
        """
        Returns the Keys
        """
        return iter(await self.aget_all_keys(True, **kwargs))
    
    async def aget_keys(self, pattern: str, exclude_base_key: Optional[bool] = None, **kwargs) -> List[str]:
        """
        Returns the Keys
        """
        raise NotImplementedError
    
    async def avalues(self, **kwargs) -> Iterable[Any]:
        """
        Returns the Values
        """
        return iter(await self.aget_all_values(**kwargs))
    
    async def aitems(self, iterable: Optional[bool] = True, **kwargs) -> Union[Iterable[ItemsView], Dict[str, Any]]:
        """
        Returns the Items
        """
        items = await self.aget_all_data(True, **kwargs)
        return items.items() if iterable else items
    
    def expire(self, key: str, ex: Optional[int] = None, **kwargs) -> None:
        """
        Expires a Key
        """
        pass
            
    async def aexpire(self, key: str, ex: Optional[int] = None, **kwargs) -> None:
        """
        Expires a Key
        """
        pass
    
    def setdefault(self, key: str, default: Any = None, update_values: Optional[bool] = False, enforce_type: Optional[bool] = False, **kwargs):
        """
        Sets a Default Value
        """
        if self.contains(key):
            with contextlib.suppress(Exception):
                value = self.get(key)
                if value is not None:
                    if enforce_type and not isinstance(value, type(default)):
                        value = type(default)(value)
                    if update_values and isinstance(value, dict) and default and isinstance(default, dict):
                        from lazyops.libs.abcs.utils.helpers import update_dict
                        value = update_dict(value, default, exclude_none = True)
                        # for k, v in default.items():
                        #     if k not in value or value[k] is None:
                        #         
                        #         value[k] = v
                        self.set(key, value)
                    return value
        self.set(key, default)
        return default
    
    async def asetdefault(self, key: str, default: Any = None, update_values: Optional[bool] = False, enforce_type: Optional[bool] = False, **kwargs):
        """
        Sets a Default Value
        """
        if await self.acontains(key):
            with contextlib.suppress(Exception):
                value = await self.aget(key)
                if value is not None:
                    if enforce_type and not isinstance(value, type(default)):
                        value = type(default)(value)
                    if update_values and isinstance(value, dict) and default and isinstance(default, dict):
                        from lazyops.libs.abcs.utils.helpers import update_dict
                        value = update_dict(value, default, exclude_none = True)
                        # for k, v in default.items():
                        #     if k not in value or value[k] is None:
                        #         value[k] = v
                        await self.aset(key, value)
                    return value
        await self.aset(key, default)
        return default
    
    def update(self, data: Dict[str, Any], **kwargs):
        """
        Updates the Cache
        """
        for key, value in data.items():
            self.set(key, value, **kwargs)
    
    async def aupdate(self, data: Dict[str, Any], **kwargs):
        """
        Updates the Cache
        """
        for key, value in data.items():
            await self.aset(key, value, **kwargs)

    def update_key(self, key: str, data: Dict[str, Any], deep: Optional[bool] = True,  exclude_none: Optional[bool] = True, **kwargs) -> Dict[str, Any]:
        """
        Updates the Dict at the Key
        """
        src = self.get(key, default = {}) or {}
        if deep:
            from lazyops.libs.abcs.utils.helpers import update_dict
            src = update_dict(src, data, exclude_none = exclude_none)
        else:
            src.update(data)
        self.set(key, src, **kwargs)
        return src


    async def aupdate_key(self, key: str, data: Dict[str, Any], deep: Optional[bool] = True,  exclude_none: Optional[bool] = True, **kwargs) -> Dict[str, Any]:
        """
        [Async] Updates the Dict at the Key
        """
        src = await self.aget(key, default = {}) or {}
        if deep:
            from lazyops.libs.abcs.utils.helpers import update_dict
            src = update_dict(src, data, exclude_none = exclude_none)
        else:
            src.update(data)
        await self.aset(key, src, **kwargs)
        return src
        
    def popitem(self, **kwargs):
        """
        Pops an Item from the Cache
        """
        self._precheck()
        key = next(iter(self))
        value = self.get(key, **kwargs)
        if value is not None:
            if self.async_enabled:
                ThreadPool.create_background_task(self.adelete(key))
            else:
                self.delete(key)
            return key, value
        return None, None
    
    async def apopitem(self, **kwargs):
        """
        Pops an Item from the Cache
        """
        self._precheck()
        key = next(iter(self))
        value = await self.aget(key, **kwargs)
        if value is not None:
            await self.adelete(key)
            return key, value
        return None, None


    def __getitem__(self, key):
        """
        Gets an Item from the DB
        """
        
        return self.get(key)

    def __setitem__(self, key, value):
        """
        Sets an Item in the Cache
        """
        return ThreadPool.create_background_task(self.aset(key, value)) if \
            self.async_enabled else self.set(key, value)
    
    def __delitem__(self, key):
        """
        Deletes an Item from the Cache
        """
        return ThreadPool.create_background_task(self.adelete(key)) if \
            self.async_enabled else self.delete(key)

    def __iter__(self):
        """
        Iterates over the Cache
        """
        return self.iterate()
    

    def __len__(self):
        """
        Returns the Length of the Cache
        """
        self._precheck()
        return len(list(self.iterate()))
    
    def length(self, **kwargs):
        """
        Returns the Length of the Cache
        """
        self._precheck()
        return len(self.get_all_keys(**kwargs))

    async def alength(self, **kwargs):
        """
        Returns the Length of the Cache
        """
        self._precheck()
        return len(await self.aget_all_keys(**kwargs))


    def __contains__(self, key):
        """
        Returns True if the Cache contains the Key
        """
        return self.contains(key)
    
    def __eq__(self, other):
        """
        Returns True if the Cache is Equal to the Other
        """
        return self.get_all_data(True) == other
    
    def __repr__(self):
        """
        Returns the Representation of the Cache
        """
        return f"<{self.__class__.__name__} num_keys={len(self)}, base_key={self.base_key}, serializer={self.serializer.name}>"
    
    def migrate_compression(self, **kwargs):
        """
        Migrates the Compression
        """
        pass

    async def amigrate_compression(self, **kwargs):
        """
        Migrates the Compression
        """
        pass

    def get_all_data_raw(self, exclude_base_key: Optional[bool] = False, **kwargs) -> Dict[str, Any]:
        """
        Loads all the Data
        """
        raise NotImplementedError
    
    async def aget_all_data_raw(self, exclude_base_key: Optional[bool] = False, **kwargs) -> Dict[str, Any]:
        """
        Loads all the Data
        """
        return await ThreadPool.run_async(self.get_all_data_raw, exclude_base_key = exclude_base_key, **kwargs)
    
    def load_data_raw(self, data: Dict[str, Any], includes_base_key: Optional[bool] = False, **kwargs):
        """
        Loads the Data
        """
        raise NotImplementedError
    
    async def aload_data_raw(self, data: Dict[str, Any], includes_base_key: Optional[bool] = False, **kwargs):
        """
        Loads the Data
        """
        return await ThreadPool.run_async(self.load_data_raw, data, includes_base_key = includes_base_key, **kwargs)
    
    def dump_data_raw(self, include_base_key: Optional[bool] = False, **kwargs) -> Dict[str, Any]:
        """
        Dumps the Data
        """
        raise NotImplementedError
    
    async def adump_data_raw(self, include_base_key: Optional[bool] = False, **kwargs) -> Dict[str, Any]:
        """
        Dumps the Data
        """
        return await ThreadPool.run_async(self.dump_data_raw, include_base_key = include_base_key, **kwargs)
    
    
    def replicate_from(self, source: Any, **kwargs):
        """
        Replicates the Cache from a Source
        """
        raise NotImplementedError
    
    async def areplicate_from(self, source: Any, **kwargs):
        """
        Replicates the Cache from a Source
        """
        return await ThreadPool.run_async(self.replicate_from, source, **kwargs)

    """
    Math Related Methods
    """

    def incr(self, key: str, amount: Union[int, float] = 1, **kwargs) -> Union[int, float]:
        """
        [int/float] Increments the value of the key by the given amount
        """
        if isinstance(amount, int): return self.incrby(key, amount = amount, **kwargs)
        return self.incrbyfloat(key, amount = amount, **kwargs)
    
    def incrby(self, key: str, amount: int = 1, **kwargs) -> int:
        """
        [int] Increments the value of the key by the given amount
        """
        raise NotImplementedError
    
    def incrbyfloat(self, key: str, amount: float = 1.0, **kwargs) -> float:
        """
        [float] Increments the value of the key by the given amount
        """
        raise NotImplementedError
    
    async def aincr(self, key: str, amount: Union[int, float] = 1, **kwargs) -> Union[int, float]:
        """
        Increments the value of the key by the given amount
        """
        if isinstance(amount, int): return await self.aincrby(key, amount = amount, **kwargs)
        return await self.aincrbyfloat(key, amount = amount, **kwargs)
    
    async def aincrby(self, key: str, amount: int = 1, **kwargs) -> int:
        """
        [int] Increments the value of the key by the given amount
        """
        raise NotImplementedError
    
    async def aincrbyfloat(self, key: str, amount: float = 1.0, **kwargs) -> float:
        """
        [float] Increments the value of the key by the given amount
        """
        raise NotImplementedError
    
    def decr(self, key: str, amount: Union[int, float] = 1, **kwargs) -> Union[int, float]:
        """
        [int/float] Decrements the value of the key by the given amount
        """
        if isinstance(amount, int): return self.decrby(key, amount = amount, **kwargs)
        return self.decrbyfloat(key, amount = amount, **kwargs)
    
    def decrby(self, key: str, amount: int = 1, **kwargs) -> int:
        """
        [int] Decrements the value of the key by the given amount
        """
        raise NotImplementedError
    
    def decrbyfloat(self, key: str, amount: float = 1.0, **kwargs) -> float:
        """
        [float] Decrements the value of the key by the given amount
        """
        raise NotImplementedError
    
    async def adecr(self, key: str, amount: Union[int, float] = 1, **kwargs) -> Union[int, float]:
        """
        Decrements the value of the key by the given amount
        """
        if isinstance(amount, int): return await self.adecrby(key, amount = amount, **kwargs)
        return await self.adecrbyfloat(key, amount = amount, **kwargs)
    
    async def adecrby(self, key: str, amount: int = 1, **kwargs) -> int:
        """
        [int] Decrements the value of the key by the given amount
        """
        raise NotImplementedError
    
    async def adecrbyfloat(self, key: str, amount: float = 1.0, **kwargs) -> float:
        """
        [float] Decrements the value of the key by the given amount
        """
        raise NotImplementedError

    """
    Set Operations
    """

    def sadd(self, key: str, *values: Any, **kwargs) -> int:
        """
        Adds the value to the set
        """
        raise NotImplementedError
        
    async def asadd(self, key: str, *values: Any, **kwargs) -> int:
        """
        Adds the value to the set
        """
        return await ThreadPool.run_async(self.sadd, key, *values, **kwargs)
    
    def slength(self, key: str, **kwargs) -> int:
        """
        Returns the length of the set
        """
        raise NotImplementedError
    
    async def aslength(self, key: str, **kwargs) -> int:
        """
        Returns the length of the set
        """
        return await ThreadPool.run_async(self.slength, key, **kwargs)
    
    
    def sismember(self, key: str, value: Any, **kwargs) -> bool:
        """
        Returns whether the value is a member of the set
        """
        raise NotImplementedError
    
    async def asismember(self, key: str, value: Any, **kwargs) -> bool:
        """
        Returns whether the value is a member of the set
        """
        return await ThreadPool.run_async(self.sismember, key, value, **kwargs)
    
    def smembers(self, key: str, **kwargs) -> List[Any]:
        """
        Returns the members of the set
        """
        raise NotImplementedError

    async def asembers(self, key: str, **kwargs) -> List[Any]:
        """
        Returns the members of the set
        """
        return await ThreadPool.run_async(self.smembers, key, **kwargs)

    def smismember(self, key: str, *values: Any, **kwargs) -> bool:
        """
        Returns whether the values are members of the set
        """
        raise NotImplementedError
    
    async def asmismember(self, key: str, *values: Any, **kwargs) -> bool:
        """
        Returns whether the values are members of the set
        """
        return await ThreadPool.run_async(self.smismember, key, *values, **kwargs)

    def srem(self, key: str, *values: Any, **kwargs) -> int:
        """
        Removes the value from the set
        """
        raise NotImplementedError
    
    async def asrem(self, key: str, *values: Any, **kwargs) -> int:
        """
        Removes the value from the set
        """
        return await ThreadPool.run_async(self.srem, key, *values, **kwargs)
    
    def spop(self, key: str, **kwargs) -> Any:
        """
        Removes and returns a random member of the set
        """
        raise NotImplementedError
    
    async def aspop(self, key: str, **kwargs) -> Any:
        """
        Removes and returns a random member of the set
        """
        return await ThreadPool.run_async(self.spop, key, **kwargs)
    
    """
    Schema Modification Methods
    """

    def migrate_schema(self, schema_map: Dict[str, Any], overwrite: Optional[bool] = None, **kwargs) -> None:
        """
        Migrates the schema
        """
        raise NotImplementedError
    
    async def amigrate_schema(self, schema_map: Dict[str, Any], overwrite: Optional[bool] = None, **kwargs) -> None:
        """
        Migrates the schema
        """
        await ThreadPool.run_async(self.migrate_schema, schema_map, overwrite = overwrite, **kwargs)

    """
    Context Manager Locks
    """
    
    def acquire_lock(self, timeout: Optional[float] = None, blocking: Optional[bool] = True, **kwargs) -> bool:
        """
        Acquires the lock
        """
        if self._lock is None: self._lock = Lock()
        try:
            return self._lock.acquire(timeout = timeout, blocking = blocking)
        except Exception as e:
            logger.error(f'Error acquiring lock: {e}')
            return False
    
    def release_lock(self):
        """
        Releases the lock
        """
        if self._lock is not None:
            try:
                self._lock.release()
            except Exception as e:
                logger.error(f'Error releasing lock: {e}')

    async def acquire_alock(self, timeout: Optional[float] = None, blocking: Optional[bool] = True, **kwargs) -> bool:
        """
        Acquires the lock
        """
        if self._alock is None: self._alock = AsyncLock()
        try:
            return await self._alock.acquire(timeout = timeout, blocking = blocking)
        except Exception as e:
            logger.error(f'Error acquiring lock: {e}')
            return False
        
    async def release_alock(self):
        """
        Releases the lock
        """
        if self._alock is not None:
            try:
                await self._alock.release()
            except Exception as e:
                logger.error(f'Error releasing lock: {e}')

    """
    Cloning Methods
    """

    def clone(
        self, 
        target: Optional[Any], 
        target_base_key: Optional[str] = None,
        schema_map: Optional[Dict[str, str]] = None,
        overwrite: Optional[bool] = False, 
        **kwargs
    ):
        """
        Clones the data from the current PersistentDict to a new PersistentDict
        """
        raise NotImplementedError

    async def aclone(
        self,
        target: Optional[Any], 
        target_base_key: Optional[str] = None,
        schema_map: Optional[Dict[str, str]] = None,
        overwrite: Optional[bool] = False, 
        excluded: Optional[Union[str, List[str]]] = None,
        **kwargs
    ):  # sourcery skip: low-code-quality
        """
        Clones the data from the target PersistentDict to a current PersistentDict
        """
        raise NotImplementedError



    def clone_from(
        self,
        target: Any, 
        target_base_key: Optional[str] = None,
        schema_map: Optional[Dict[str, str]] = None,
        overwrite: Optional[bool] = False, 
        excluded: Optional[Union[str, List[str]]] = None,
        **kwargs
    ): 
        """
        Clones the data from the target PersistentDict to a current PersistentDict
        """
        raise NotImplementedError


    async def aclone_from(
        self,
        target: Any, 
        target_base_key: Optional[str] = None,
        schema_map: Optional[Dict[str, str]] = None,
        overwrite: Optional[bool] = False, 
        excluded: Optional[Union[str, List[str]]] = None,
        **kwargs
    ): 
        """
        Clones the data from the target PersistentDict to a current PersistentDict
        """
        raise NotImplemented



    """
    Backend Supported Methods
    """

    def select(self, *args, **kwargs) -> Any:
        """
        Selects the data from the backend
        """
        raise NotImplementedError
    
    async def aselect(self, *args, **kwargs) -> Any:
        """
        [Async] Selects the data from the backend
        """
        raise NotImplementedError


"""
Google OAuth2 Provider
"""
from __future__ import annotations

from .config import GoogleOAuth2Config
from .client import GoogleOAuth2Client
from .token import (
    IDToken as GoogleIDToken,
    AccessToken as GoogleAccessToken,
    OAuth2Token as GoogleOAuth2Token,
)


from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    
    TokenT = Union[GoogleAccessToken, GoogleIDToken, GoogleOAuth2Token]



## Base Object Models
from .chat import ChatMessage, ChatChoice, ChatObject, ChatResponse, Function, FunctionCall, Tool, JSONSchema
from .completions import CompletionChoice, CompletionObject, CompletionResponse
from .embeddings import EmbeddingData, EmbeddingObject, EmbeddingResponse
from .rerankings import RerankingData, RerankingObject, RerankingResponse
from .models import ModelData, ModelObject, ModelResponse
from .functions import BaseFunctionModel, BaseFunction

## Route Models
from .chat import ChatRoute
from .completions import CompletionRoute
from .embeddings import EmbeddingRoute
from .rerankings import RerankingRoute
from .models import ModelRoute





from __future__ import annotations

"""Helpers for maintaining stateful KVDB client registries."""

import abc
import copy
import typing as t

if t.TYPE_CHECKING:
    from kvdb import KVDBSession, PersistentDict

__all__ = ['KVDBRegistry']


class KVDBRegistry(abc.ABC):
    """Manage shared KVDB sessions and persistent dictionaries for a module."""

    module: t.Optional[str] = None

    default_serializer: t.Optional[str] = 'json'
    default_kwargs: t.Dict[str, t.Any] = {}

    default_pdict_expiration: t.Optional[int] = None
    default_pdict_kwargs: t.Dict[str, t.Any] = {}

    # If ``None`` we fall back to ``default_serializer``; if ``'none'`` we skip serialisation
    default_fdict_serializer: t.Optional[str] = None
    default_fdict_expiration: t.Optional[int] = -1
    default_fdict_kwargs: t.Dict[str, t.Any] = {}

    sessions: t.Dict[str, 'KVDBSession'] = {}
    pdicts: t.Dict[str, 'PersistentDict'] = {}
    pdict_aliases: t.Dict[str, str] = {}
    pdict_prefix_module: t.Optional[bool] = True

    _extra: t.Dict[str, t.Any] = {}

    def __init__(
        self,
        module: t.Optional[str] = None,
        default_serializer: t.Optional[str] = None,
        default_kwargs: t.Optional[t.Mapping[str, t.Any]] = None,
        default_pdict_expiration: t.Optional[int] = None,
        default_pdict_kwargs: t.Optional[t.Mapping[str, t.Any]] = None,
        default_fdict_serializer: t.Optional[str] = None,
        default_fdict_expiration: t.Optional[int] = None,
        default_fdict_kwargs: t.Optional[t.Mapping[str, t.Any]] = None,
        pdict_prefix_module: t.Optional[bool] = None,
        **kwargs: t.Any,
    ) -> None:
        """Configure default serialisation options for KVDB resources."""

        self._extra = {}
        if module:
            self.module = module
        if default_serializer:
            self.default_serializer = default_serializer
        if default_kwargs:
            self.default_kwargs = dict(default_kwargs)
        if default_pdict_expiration:
            self.default_pdict_expiration = default_pdict_expiration
        if default_pdict_kwargs:
            self.default_pdict_kwargs = dict(default_pdict_kwargs)
        if pdict_prefix_module is not None:
            self.pdict_prefix_module = pdict_prefix_module
        if default_fdict_serializer:
            self.default_fdict_serializer = default_fdict_serializer
        elif self.default_serializer is None:
            self.default_fdict_serializer = self.default_serializer
        if default_fdict_expiration:
            self.default_fdict_expiration = default_fdict_expiration
        if default_fdict_kwargs:
            self.default_fdict_kwargs = dict(default_fdict_kwargs)

        if not self.module:
            self.pdict_prefix_module = False
        self.post_init(**kwargs)

    def post_init(self, **kwargs: t.Any) -> None:  # pragma: no cover - hook for subclasses
        """Finalise subclass configuration after initial settings are applied."""

    @property
    def module_prefix(self) -> t.Optional[str]:
        """Return the prefix applied to session and persistence names."""

        return self.module

    def _build_kwargs(self, src_kwargs: t.Mapping[str, t.Any], **kwargs: t.Any) -> t.Dict[str, t.Any]:
        """Clone ``src_kwargs`` and merge ``kwargs`` without mutating inputs."""

        merged = copy.deepcopy(src_kwargs)
        if kwargs:
            merged.update(kwargs)
        return merged

    def get(
        self,
        name: t.Optional[str] = None,
        serializer: t.Optional[str] = None,
        **kwargs: t.Any,
    ) -> 'KVDBSession':
        """Fetch or create a named KVDB session."""

        session_name = name or self.module_prefix or 'global'
        if session_name not in self.sessions:
            from kvdb import KVDBClient

            effective_serializer = serializer or self.default_serializer
            if effective_serializer == 'none':
                effective_serializer = None
            if self.default_kwargs:
                kwargs = self._build_kwargs(self.default_kwargs, **kwargs)
            self.sessions[session_name] = KVDBClient.get_session(
                name=session_name,
                serializer=effective_serializer,
                **kwargs,
            )
        return self.sessions[session_name]

    def pdict(
        self,
        base_key: str,
        expiration: t.Optional[int] = None,
        aliases: t.Optional[t.Iterable[str]] = None,
        **kwargs: t.Any,
    ) -> 'PersistentDict':
        """Return a persistent dictionary scoped to ``base_key``.

        Args:
            base_key: Identifier for the persistence bucket.
            expiration: Optional TTL applied during creation. ``<=0`` disables.
            aliases: Alternate keys that reference the same dictionary.
            **kwargs: Additional arguments forwarded to ``create_persistence``.
        """

        effective_key = base_key
        if self.pdict_prefix_module and self.module_prefix and self.module_prefix not in effective_key:
            effective_key = f'{self.module_prefix}.{effective_key}'

        if effective_key not in self.pdicts and effective_key not in self.pdict_aliases:
            persistence_session = self.get(
                f'{self.module_prefix}.persistence' if self.module_prefix else 'persistence',
                serializer='none',
                url=kwargs.pop('url', None),
            )
            ttl = expiration
            if ttl is None and self.default_pdict_expiration:
                ttl = self.default_pdict_expiration
            if ttl is not None and ttl <= 0:
                ttl = None
            if self.default_pdict_kwargs:
                kwargs = self._build_kwargs(self.default_pdict_kwargs, **kwargs)
            self.pdicts[effective_key] = persistence_session.create_persistence(
                base_key=effective_key,
                expiration=ttl,
                **kwargs,
            )
            if aliases:
                for alias in aliases:
                    self.pdict_aliases[alias] = effective_key
        elif effective_key in self.pdict_aliases:
            effective_key = self.pdict_aliases[effective_key]

        return self.pdicts[effective_key]



from __future__ import annotations

"""
FastAPI Extensions: OAuth2 Manager
"""

import abc
from lzl.types import eproperty
from lzl.logging import logger, null_logger, Logger
from lzl import load
from lzo.utils.helpers.base import create_unique_secret_key
from starlette.authentication import AuthenticationError
from lzl.ext.fast.types.security import Authorization, APIKey, AuthzKey
from urllib.parse import urljoin
from .types import errors
from .types.user import AuthMethod
# from .types.security import HTTPBase, HTTPAuthorizationCredentials, Authorization
from typing import Optional, Dict, Any, List, Set, Union, Type, Tuple, Callable, Awaitable, TYPE_CHECKING

if TYPE_CHECKING:
    from fastapi import FastAPI
    from .config import OAuth2Config
    from .providers import ProviderClientT
    from .types import Request
    from .types.user import OAuth2User, UnauthenticatedUser, APIKeyUser
    from .types.token import OAuth2Token
    from .types.security import HTTPBase, HTTPAuthorizationCredentials
    from .middleware import OAuth2Middleware
    from .backend import OAuth2Backend
    from starlette.types import Scope
else:
    from fastapi import Request

if load.TYPE_CHECKING:
    import jwt
else:
    jwt = load.LazyLoad("jwt", install_missing=True, install_options={'package': 'pyjwt'})

_DefaultRoutes: Dict[str, Any] = {
    'login': '/login',
    'start': '/start',
    'callback': '/callback',
    'logout': '/logout',
    'apikey': '/apikey',
    'validate': '/auth/validate',
    'user_info': '/userinfo',
    'default_redirect': '/docs',
}

class OAuth2Manager(abc.ABC):
    """
    The OAuth2 Manager
    """
    _extra: Dict[str, Any] = {}

    def __init__(
        self,
        config: 'OAuth2Config',
        **kwargs,
    ) -> None:
        """
        Initializes the OAuth2 Manager
        """
        self.config = config
        self.providers: Dict[str, 'ProviderClientT'] = {}
        self.jwt_issuers: Dict[str, str] = {} # A proxy for jwt_issuer_url and provider name
        self.default_provider: Optional[str] = None
        self.routes: Dict[str, Any] = _DefaultRoutes.copy()
        self.custom_auth_handlers: List[Callable[..., Awaitable[Tuple[bool, bool, bool, Tuple[Any, Any, Any]]]]] = []
        self.custom_middleware_handlers: List[Callable[..., Awaitable['Scope']]] = []
        self.optional_paths: Set[str] = {
            path for route, path in self.routes.items() if route != 'user_info'
        }
        self.excluded_paths: Set[str] = set()
        self.post_init(**kwargs)
    
    def post_init(self, **kwargs):
        """
        Post Init
        """
        pass

    def add_optional_path(self, path: Union[str, List[str]]):
        """
        Adds an optional path
        """
        if isinstance(path, str): path = [path]
        self.optional_paths.update(path)

    def add_excluded_path(self, path: Union[str, List[str]]):
        """
        Adds an excluded path
        """
        if isinstance(path, str): path = [path]
        self.excluded_paths.update(path)
        if self.oauth2_middleware is not None:
            self.oauth2_middleware.excluded_url_paths.update(path)

    def optional(self, path: str) -> str:
        """
        Adds an optional path
        """    
        self.add_optional_path(path)
        return path

    def exclude(self, path: str) -> str:
        """
        Adds an excluded path
        """    
        self.add_excluded_path(path)
        return path
    
    def add_auth_handler(self, handler: Callable[..., Awaitable[Tuple[bool, bool, bool, Tuple[Any, Any, Any]]]]):
        """
        Adds an auth handler

        The handler should be a function that takes the following arguments:
        - request: The request object
        - optional: Whether the auth is optional
        - kwargs: Any additional kwargs

        It should return a tuple of tuple (valid, set_state, should_continue, ((token, user, apikey) or None))
        """
        self.custom_auth_handlers.append(handler)

    def add_middleware_handler(self, handler: Callable[..., Awaitable['Scope']]):
        """
        Adds a middleware handler
        """
        self.custom_middleware_handlers.append(handler)
    
    @property
    def user_class(self) -> Type['OAuth2User']:
        """
        Returns the User Class
        """
        if 'user_class' not in self._extra:
            from .types.user import OAuth2User
            self._extra['user_class'] = OAuth2User
        return self._extra['user_class']
    
    @user_class.setter
    def user_class(self, value: Union[Type['OAuth2User'], str]) -> None:
        """
        Sets the User Class
        """
        if isinstance(value, str): value = load.lazy_import(value)
        self._extra['user_class'] = value

    @property
    def unauth_user_class(self) -> Type['UnauthenticatedUser']:
        """
        Returns the Unauthenticated User Class
        """
        if 'unauth_user_class' not in self._extra:
            from .types.user import UnauthenticatedUser
            self._extra['unauth_user_class'] = UnauthenticatedUser
        return self._extra['unauth_user_class']
    
    @unauth_user_class.setter
    def unauth_user_class(self, value: Union[Type['UnauthenticatedUser'], str]) -> None:
        """
        Sets the Unauthenticated User Class
        """
        if isinstance(value, str): value = load.lazy_import(value)
        self._extra['unauth_user_class'] = value

    @property
    def provider_pattern(self) -> str:
        """
        Returns the provider pattern

        '^(google|kinde|auth0)$'
        """
        # base_pattern = '|'.join(list(self.providers.keys()))
        # logger.info(f'Provider Pattern: {base_pattern}')
        # return f'^({base_pattern})$'
        if 'provider_pattern' not in self._extra:
            base_pattern = '|'.join(list(self.providers.keys()))
            self._extra['provider_pattern'] = f'^({base_pattern})$'
        return self._extra['provider_pattern']

    @property
    def app(self) -> 'FastAPI':
        """
        Returns the app
        """
        return self.config.app
    
    @eproperty
    def oauth2_middleware(self) -> 'OAuth2Middleware':
        """
        Returns the OAuth2 Middleware
        """
        return self._extra.get('oauth2_middleware')
    
    @eproperty
    def oauth2_backend(self) -> 'OAuth2Backend':
        """
        Returns the OAuth2 Backend
        """
        return self._extra.get('oauth2_backend')
    
    @property
    def state_kw_map(self) -> Dict[str, str]:
        """
        Returns the state keyword map
        """
        return self.config.session_state_kws
    
    @eproperty
    def session_state_kw_check(self) -> str:
        """
        Returns the first key in the state keyword map to validate
        whether there is a state present in the session
        """
        return list(self.config.session_state_kws.values())[0]
    
    def fetch_auth_from_state(self, request: 'Request') -> Tuple['OAuth2Token', 'OAuth2User', Optional[str]]:
        """
        Fetches the auth from the state
        """
        token = request.state._state.get(self.state_kw_map['token'])
        user = request.state._state.get(self.state_kw_map['user'])
        apikey = request.state._state.get(self.state_kw_map['apikey'])
        return (token, user, apikey)
    
    @eproperty
    def session_oauth2_key(self) -> str:
        """
        Returns the session oauth2 key
        """
        return self.config.session_oauth2_key
    
    @property
    def domains(self) -> List[str]:
        """
        Returns the domains
        """
        return self.config.domains
    
    @property
    def disabled_paths(self) -> List[str]:
        """
        Returns the disabled paths
        """
        return self.config.disabled_paths
    
    @property
    def api_key_users(self) -> Dict[str, 'APIKeyUser']:
        """
        Returns the API Key Users
        """
        return self.config.api_key_users
    
    @property
    def autologger(self) -> 'Logger':
        """
        Returns the autologger
        """
        return logger if self.config.debug_enabled else null_logger

    def update_routes(self, routes: Dict[str, Union[str, Dict[str, Any]]]) -> None:
        """
        Updates the routes
        """
        for name, route in routes.items():
            if isinstance(route, str): self.routes[name] = route
            elif isinstance(route, dict): 
                if route.get('disabled') is True: 
                    if name in self.routes: del self.routes[name]
                    continue
                self.routes[name] = route.get('path', name)

    def load_providers(
        self, 
        providers: List[Dict[str, Any]], 
    ) -> None:
        """
        Loads the providers
        """
        from .providers import get_provider_classes
        for config in providers:
            provider = config['name']
            enabled = config.pop('enabled', True)
            if not enabled: continue
            is_default = config.pop('default', False)
            _config_cls, _client_cls = get_provider_classes(
                provider,
                config_class = config.pop('config_class', None),
                client_class = config.pop('client_class', None),
            )
            _config = _config_cls(**config)
            self.providers[provider] = _client_cls(
                manager = self,
                config = _config,
            )
            if _config.jwt_issuer_url: self.jwt_issuers[_config.jwt_issuer_url] = provider
            if is_default: self.default_provider = provider
        if not self.default_provider: self.default_provider = list(self.providers.keys())[0]
        if not self.config.temp_data.has_logged('providers_loaded'):
            logger.info(f'Loaded |g|{len(self.providers)}|e| OAuth2 Providers: |y|{list(self.providers.keys())}|e| - Default Provider: |g|{self.default_provider}|e|', colored = True)

    def get_oauth_provider(self, provider: str, raise_errors: Optional[bool] = True) -> 'ProviderClientT':
        """
        Returns the OAuth2 provider
        """
        if provider not in self.providers:
            if raise_errors: raise errors.InvalidProviderError(detail=f'Invalid Provider: {provider}')
            return None
        return self.providers[provider]


    # def get_auth_method_from_request(self, request: 'Request') -> Optional[AuthMethod]:

    async def get_provider_for_jwt(
        self,
        token: str,
        request: 'Request',
    ) -> 'ProviderClientT':
        """
        Resolves the provider for the jwt
        """
        # Special Case for Google
        if token.startswith('ya29.') and 'google' in self.providers:
            return self.providers['google']
        decoded: Dict[str, Union[str, List[str]]] = jwt.decode(
            token,
            algorithms = ['RS256'],
            options={'verify_signature': False}
        )
        if decoded.get('iss') not in self.jwt_issuers:
            raise errors.InvalidProviderError(provider = decoded.get('iss'))
        return self.providers[self.jwt_issuers[decoded.get('iss')]]
    
    def decrypt_api_key(self, key: str) -> Dict[str, Any]:
        """
        Decrypts the API Key
        """
        return self.config.decrypt_api_key_data(key)
    
    def encrypt_api_key(self, data: Dict[str, Any]) -> str:
        """
        Encrypts the API Key
        """
        return self.config.encrypt_api_key_data(data)


    """
    FastAPI Handler Methods
    """


    """
    FastAPI Routes
    """

    def fetch_app_endpoint_from_request(self, request: Request) -> str:
        """
        Fetches the app endpoint from the request
        """
        if request.session.get('app_endpoint'):
            return request.session['app_endpoint']
        headers = dict(request.headers.items())
        if headers.get(':authority:'):
            return headers[':scheme:'] + '://' + headers[':authority:']
        if headers.get('x-forwarded-host'):
            if headers.get('x-forwaded-scheme'):
                return headers['x-forwaded-scheme'] + '://' + headers['x-forwarded-host']
            if self.config.in_k8s:
                return 'https://' + headers['x-forwarded-host']
            forwaded_port = headers['x-forwarded-port']
            if '443' in forwaded_port: return 'https://' + headers['x-forwarded-host']
            return 'http://' + headers['x-forwarded-host']
        if headers.get('host'):
            scheme = 'https' if self.config.in_k8s else 'http'
            return f'{scheme}://' + headers['host']
        raise ValueError('Unable to determine app endpoint')


    def mount_components(
        self,
        app: Optional['FastAPI'] = None,
        include_in_schema: Optional[bool] = None,
        disable_middleware: Optional[bool] = None,
        disable_session: Optional[bool] = None,
        disable_routes: Optional[bool] = None,
        **kwargs,
    ):
        """
        Mounts all the components
        """
        if app is not None: self.config.app = app
        elif self.config.app is not None: app = self.config.app
        else: raise ValueError('app must be provided')

        # add the middlewares
        if not disable_middleware:
            # logger.info(f'Adding OAuth2 Middleware: {self.config.middleware_class}')
            app.add_middleware(
                self.config.middleware_class,
                config = self.config,
                backend_class = self.config.backend_class,
            )

        disable_session = disable_session or self.config.session_disabled
        if not disable_session:
            # logger.info(f'Adding OAuth2 Session Middleware: {self.config}')
            state_kws = list(self.config.session_state_kws.values())
            app.add_middleware(
                self.config.session_class,
                config = self.config,
                oauth2_enabled = True,
                state_kws = state_kws,
                **self.config.session_config,
            )
        
        if not disable_routes:
            self.mount_oauth_routes(
                app = app,
                include_in_schema = include_in_schema,
                **kwargs,
            )


    def mount_oauth_routes(
        self,
        app: 'FastAPI',
        include_in_schema: Optional[bool] = None,
        **kwargs,
    ):
        """
        Mounts the OAuth2 Routes
        """
        from fastapi import APIRouter,  Path, Query
        from fastapi.responses import RedirectResponse
        from lzl.ext.fast.types.responses import PrettyJSONResponse
        router = APIRouter(
            include_in_schema = include_in_schema, 
            default_response_class = PrettyJSONResponse
        )

        if 'login' in self.routes:
            if self.config.api_keys_enabled:
                @router.get(self.routes['login'])
                async def login(
                    request: Request,
                    auth: Authorization,
                    apikey: APIKey,
                    authzkey: AuthzKey,
                    redirect: str = Query(None, description='Redirect URL'),
                ):
                    """
                    Login Endpoint
                    """
                    headers = {}
                    if request.user.is_authenticated: 
                        headers[self.config.header] = f'{self.config.scheme} {request.auth.value}'
                        if redirect: return RedirectResponse(url = redirect, headers = headers)
                        if redirect := request.session.pop('oauth2_redirect', None):
                            return RedirectResponse(url = redirect, headers = headers)
                        if self.routes.get('default_redirect'):
                            return RedirectResponse(url = self.routes['default_redirect'], headers = headers)
                        return {'login': 'success'}
                    provider_url = self.routes['start'] + "/" + self.default_provider
                    return RedirectResponse(url = provider_url)

            else:
                @router.get(self.routes['login'])
                async def login(
                    request: Request,
                    auth: Authorization,
                    redirect: str = Query(None, description='Redirect URL'),
                ):
                    """
                    Login Endpoint
                    """
                    headers = {}
                    if request.user.is_authenticated: 
                        headers[self.config.header] = f'{self.config.scheme} {request.auth.value}'
                        if redirect: return RedirectResponse(url = redirect, headers = headers)
                        if redirect := request.session.pop('oauth2_redirect', None):
                            return RedirectResponse(url = redirect, headers = headers)
                        if self.routes.get('default_redirect'):
                            return RedirectResponse(url = self.routes['default_redirect'], headers = headers)
                        return {'login': 'success'}
                    provider_url = self.routes['start'] + "/" + self.default_provider
                    return RedirectResponse(url = provider_url)

        if 'start' in self.routes:
            @router.get(self.routes['start'] + "/{provider}")
            async def start_with_provider(
                request: Request,
                provider: str = Path(..., description='OAuth2 Provider', pattern = self.provider_pattern),
                redirect: str = Query(None, description='Redirect URL'),
                scope: str = Query(None, description='Scope'),
            ):
                """
                Authorize Endpoint with an OAuth2 Provider
                """
                if redirect: request.session['oauth2_redirect'] = redirect
                state_id = create_unique_secret_key()
                request.session['state_id'] = state_id
                client = self.get_oauth_provider(provider)
                provider_cb_path = self.routes['callback'] + "/" + provider
                provider_redirect_url = urljoin(self.fetch_app_endpoint_from_request(request), provider_cb_path)
                auth_redirect_url = client.get_auth_url(
                    redirect_url = provider_redirect_url,
                    scope = scope,
                    state = state_id,
                )
                return RedirectResponse(auth_redirect_url)
        
        if 'callback' in self.routes:
            @router.get(self.routes['callback'] + "/{provider}")
            async def callback_with_provider(
                request: Request,
                provider: str = Path(..., description='OAuth2 Provider', pattern = self.provider_pattern),
                code: str = Query(..., description='Authorization Code'),
                state: str = Query(None, description='State'),
            ):
                """
                The Provider Callback Endpoint
                """
                client = self.get_oauth_provider(provider)
                provider_cb_path = self.routes['callback'] + "/" + provider
                provider_redirect_url = urljoin(self.fetch_app_endpoint_from_request(request), provider_cb_path)
                try:
                    token = await client.afetch_token(
                        code = code,
                        redirect_url = provider_redirect_url,
                        request = request,
                    )
                except Exception as e:
                    logger.error(f'Error fetching token: {e}')
                    raise errors.InvalidAuthorizationError(error = e) from e
                self.autologger.info(f'Fetched Token: |g|{token}|e|', colored = True, prefix = provider)
                oauth_token = await client.aparse_token_from_callback(token)
                self.autologger.info(f'Fetched Token: |g|{oauth_token}|e|', colored = True, prefix = provider)
                request.session[self.session_oauth2_key] = {
                    'provider': provider,
                    'identity': oauth_token.identity,
                }
                redirect = request.session.get('oauth2_redirect')
                if not redirect and self.routes.get('default_redirect'): redirect = self.routes['default_redirect']
                headers = {
                    self.config.header: f'{self.config.scheme} {oauth_token.value}'
                }
                if redirect: response = RedirectResponse(url = redirect, headers = headers)
                else: response = PrettyJSONResponse({'login': 'success'}, headers = headers)
                return response

        if 'logout' in self.routes:
            @router.get(self.routes['logout'])
            async def logout(
                request: Request,
                redirect: str = Query(None, description='Redirect URL'),
            ):
                """
                Logout Endpoint
                """
                if request.user.is_authenticated:
                    await request.user.alogout()
                if not redirect: redirect = request.session.get('oauth2_redirect')
                request.session.clear()
                request.state._state.clear()
                if redirect: return RedirectResponse(url = redirect)
                if self.routes.get('default_redirect'):
                    return RedirectResponse(url = self.routes['default_redirect'])
                return {'logout': 'success'}
        
        if 'user_info' in self.routes:
            if self.config.api_keys_enabled:
                @router.get(self.routes['user_info'])
                async def user_info(
                    request: Request,
                    auth: Authorization,
                    apikey: APIKey,
                    authzkey: AuthzKey,
                ):
                    """
                    User Info Endpoint
                    """
                    if not request.user.is_authenticated: return {'user_info': 'not_authenticated'}
                    return PrettyJSONResponse(
                        content = {
                            'authenticated': True,
                            **request.user.get_user_info()
                        }
                    )
            else:
                @router.get(self.routes['user_info'])
                async def user_info(
                    request: Request,
                    auth: Authorization,
                ):
                    """
                    User Info Endpoint
                    """
                    if not request.user.is_authenticated: return {'user_info': 'not_authenticated'}
                    return PrettyJSONResponse(
                        content = {
                            'authenticated': True,
                            **request.user.get_user_info()
                        }
                    )
        
        if 'apikey' in self.routes:
            @router.get(self.routes['apikey'], include_in_schema = include_in_schema)
            async def get_apikey(
                request: Request,
                redirect: str = Query(None, description='Redirect URL'),
            ):
                """
                API Key Endpoint
                """
                if not request.user.is_authenticated: 
                    raise errors.InvalidAuthorizationError('No Authorization')
                headers = {'x-authz-key': request.user.apikey}
                if redirect: return RedirectResponse(url = redirect, headers = headers)
                return PrettyJSONResponse(
                    content = {
                        'apikey': request.user.apikey,
                    },
                    headers = headers,
                )

        # Clean these up later
        if 'validate' in self.routes:
            if self.config.api_keys_enabled:
                @router.get(self.routes['validate'], include_in_schema = include_in_schema)
                async def validate_auth(
                    request: Request,
                    auth: Authorization,
                    apikey: APIKey,
                    authzkey: AuthzKey,
                    redirect: str = Query(None, description='Redirect URL'),
                ):
                    """
                    Validate Endpoint
                    """
                    if not request.user.is_authenticated: 
                        raise errors.InvalidAuthorizationError('No Authorization')
                    headers = {'x-authz-key': request.user.apikey}
                    if redirect: return RedirectResponse(url = redirect, headers = headers)
                    return PrettyJSONResponse(
                        content = {
                            'authenticated': True,
                        },
                        headers = headers,
                    )
            else:
                @router.get(self.routes['validate'], include_in_schema = include_in_schema)
                async def validate_auth(
                    request: Request,
                    auth: Authorization,
                    redirect: str = Query(None, description='Redirect URL'),
                ):
                    """
                    Validate Endpoint
                    """
                    if not request.user.is_authenticated: 
                        raise errors.InvalidAuthorizationError('No Authorization')
                    headers = {'x-authz-key': request.user.apikey}
                    if redirect: return RedirectResponse(url = redirect, headers = headers)
                    return PrettyJSONResponse(
                        content = {
                            'authenticated': True,
                        },
                        headers = headers,
                    )

        app.include_router(router, tags = ['OAuth2'])

    @classmethod
    def load(
        cls, 
        app: Optional['FastAPI'] = None, 
        config: Optional['OAuth2Config'] = None,
        **kwargs
    ) -> 'OAuth2Manager':
        """
        Loads the OAuth2 Manager
        """
        if config is None:
            from .config import OAuth2Config
            config = OAuth2Config(manager_class = cls)

        if app is not None: config.app = app
        if kwargs: config.update_config(**kwargs)
        return config.get_manager()
